@@ -2,9 +2,10 @@
 
 import functools
 import io
-import shutil
+from hashlib import md5
+from pathlib import Path
 
-from . import CSS
+from . import CSS, DEFAULT_OPTIONS
 from .anchors import gather_anchors, make_page_bookmark_tree
 from .css import get_all_computed_styles
 from .css.counters import CounterStyle
@@ -16,7 +17,8 @@ from .images import get_image_from_uri as original_get_image_from_uri
 from .layout import LayoutContext, layout_document
 from .logger import PROGRESS_LOGGER
 from .matrix import Matrix
-from .pdf import generate_pdf
+from .pdf import VARIANTS, generate_pdf
+from .pdf.metadata import generate_rdf_metadata
 from .text.fonts import FontConfiguration
 
 
@@ -27,6 +29,7 @@ class Page:
     instantiated directly.
 
     """
+
     def __init__(self, page_box):
         #: The page width, including margins, in CSS pixels.
         self.width = page_box.margin_width()
@@ -66,39 +69,21 @@ class Page:
         #: ``(x, y)`` point in CSS pixels from the top-left of the page.
         self.anchors = {}
 
-        #: The :obj:`list` of ``(element, attributes, rectangle)`` :obj:`tuples
-        #: <tuple>`. A ``rectangle`` is ``(x, y, width, height)``, in CSS
+        #: The :obj:`dict` mapping form elements to a list
+        #: of ``(element, attributes, rectangle)`` :obj:`tuples <tuple>`.
+        #: A ``rectangle`` is ``(x, y, width, height)``, in CSS
         #: pixels from the top-left of the page. ``atributes`` is a
         #: :obj:`dict` of HTML tag attributes and values.
-        self.inputs = []
+        #: The key ``None`` will contain inputs that are not part of a form.
+        self.forms = {None: []}
 
-        gather_anchors(
-            page_box, self.anchors, self.links, self.bookmarks, self.inputs)
+        gather_anchors(page_box, self.anchors, self.links, self.bookmarks, self.forms)
         self._page_box = page_box
 
-    def paint(self, stream, left_x=0, top_y=0, scale=1, clip=False):
-        """Paint the page into the PDF file.
-
-        :type stream: ``document.Stream``
-        :param stream:
-            A document stream.
-        :param float left_x:
-            X coordinate of the left of the page, in PDF points.
-        :param float top_y:
-            Y coordinate of the top of the page, in PDF points.
-        :param float scale:
-            Zoom scale.
-        :param bool clip:
-            Whether to clip/cut content outside the page. If false or
-            not provided, content can overflow.
-
-        """
+    def paint(self, stream, scale=1):
+        """Paint the page into the PDF file."""
         with stacked(stream):
-            # Make (0, 0) the top-left corner, and make user units CSS pixels:
-            stream.transform(a=scale, d=scale, e=left_x, f=top_y)
-            if clip:
-                stream.rectangle(0, 0, self.width, self.height)
-                stream.clip()
+            stream.transform(a=scale, d=scale)
             draw_page(self._page_box, stream)
 
 
@@ -106,11 +91,10 @@ class DocumentMetadata:
     """Meta-information belonging to a whole :class:`Document`.
 
     New attributes may be added in future versions of WeasyPrint.
-
     """
-    def __init__(self, title=None, authors=None, description=None,
-                 keywords=None, generator=None, created=None, modified=None,
-                 attachments=None, lang=None, custom=None):
+    def __init__(self, title=None, authors= None, description=None, keywords=None,
+                 generator=None, created=None, modified=None, attachments=None,
+                 lang=None, custom=None, generate_rdf_metadata=generate_rdf_metadata):
         #: The title of the document, as a string or :obj:`None`.
         #: Extracted from the ``<title>`` element in HTML
         #: and written to the ``/Title`` info field in PDF.
@@ -146,8 +130,7 @@ class DocumentMetadata:
         #: Extracted from the ``<meta name=dcterms.modified>`` element in HTML
         #: and written to the ``/ModDate`` info field in PDF.
         self.modified = modified
-        #: File attachments, as a list of tuples of URL and a description or
-        #: :obj:`None`. (Defaults to the empty list.)
+        #: A list of :class:`attachments <weasyprint.Attachment>`, empty by default.
         #: Extracted from the ``<link rel=attachment>`` elements in HTML
         #: and written to the ``/EmbeddedFiles`` dictionary in PDF.
         self.attachments = attachments or []
@@ -157,6 +140,56 @@ class DocumentMetadata:
         #: Custom metadata, as a dict whose keys are the metadata names and
         #: values are the metadata values.
         self.custom = custom or {}
+        #: Custom RDF metadata generator, which will replace the default generator.
+        #: The function should return bytes containing an RDF XML.
+        self.generate_rdf_metadata = generate_rdf_metadata
+
+
+class DiskCache:
+    """Dict-like storing images content on disk.
+
+    Bytestring values are stored on disk. Other lightweight Python objects
+    (i.e. RasterImage instances) are still stored in memory.
+
+    """
+
+    def __init__(self, folder):
+        self._path = Path(folder)
+        self._path.mkdir(parents=True, exist_ok=True)
+        self._memory_cache = {}
+        self._disk_paths = set()
+
+    def _path_from_key(self, key):
+        digest = md5(key.encode(), usedforsecurity=False).hexdigest()
+        return self._path / digest
+
+    def __getitem__(self, key):
+        if key in self._memory_cache:
+            return self._memory_cache[key]
+        else:
+            return self._path_from_key(key).read_bytes()
+
+    def __setitem__(self, key, value):
+        if isinstance(value, bytes):
+            path = self._path_from_key(key)
+            self._disk_paths.add(path)
+            path.write_bytes(value)
+        else:
+            self._memory_cache[key] = value
+
+    def __contains__(self, key):
+        return (
+            key in self._memory_cache or
+            self._path_from_key(key).exists())
+
+    def __del__(self):
+        try:
+            for path in self._disk_paths:
+                path.unlink(missing_ok=True)
+            self._path.rmdir()
+        except Exception:
+            # Silently ignore errors while clearing cache
+            pass
 
 
 class Document:
@@ -171,29 +204,28 @@ class Document:
     """
 
     @classmethod
-    def _build_layout_context(cls, html, stylesheets, presentational_hints,
-                              optimize_size, font_config, counter_style,
-                              image_cache, forms):
-        if font_config is None:
-            font_config = FontConfiguration()
-        if counter_style is None:
-            counter_style = CounterStyle()
+    def _build_layout_context(cls, html, font_config, counter_style, options):
         target_collector = TargetCollector()
         page_rules = []
         user_stylesheets = []
-        image_cache = {} if image_cache is None else image_cache
-        for css in stylesheets or []:
+        cache = options['cache']
+        if cache is None:
+            cache = {}
+        elif not isinstance(cache, (dict, DiskCache)):
+            cache = DiskCache(cache)
+        for css in options['stylesheets'] or []:
             if not hasattr(css, 'matcher'):
                 css = CSS(
                     guess=css, media_type=html.media_type,
                     font_config=font_config, counter_style=counter_style)
             user_stylesheets.append(css)
         style_for = get_all_computed_styles(
-            html, user_stylesheets, presentational_hints, font_config,
-            counter_style, page_rules, target_collector, forms)
+            html, user_stylesheets, options['presentational_hints'],
+            font_config, counter_style, page_rules, target_collector,
+            options['pdf_forms'])
         get_image_from_uri = functools.partial(
-            original_get_image_from_uri, cache=image_cache,
-            url_fetcher=html.url_fetcher, optimize_size=optimize_size)
+            original_get_image_from_uri, cache=cache,
+            url_fetcher=html.url_fetcher, options=options)
         PROGRESS_LOGGER.info('Step 4 - Creating formatting structure')
         context = LayoutContext(
             style_for, get_image_from_uri, font_config, counter_style,
@@ -201,8 +233,7 @@ class Document:
         return context
 
     @classmethod
-    def _render(cls, html, stylesheets, presentational_hints, optimize_size,
-                font_config, counter_style, image_cache, forms):
+    def _render(cls, html, font_config, counter_style, options):
         if font_config is None:
             font_config = FontConfiguration()
 
@@ -210,8 +241,7 @@ class Document:
             counter_style = CounterStyle()
 
         context = cls._build_layout_context(
-            html, stylesheets, presentational_hints, optimize_size,
-            font_config, counter_style, image_cache, forms)
+            html, font_config, counter_style, options)
 
         root_box = build_formatting_structure(
             html.etree_element, context.style_for, context.get_image_from_uri,
@@ -222,12 +252,11 @@ class Document:
         rendering = cls(
             [Page(page_box) for page_box in page_boxes],
             DocumentMetadata(**get_html_metadata(html)),
-            html.url_fetcher, font_config, optimize_size)
+            html.url_fetcher, font_config)
         rendering._html = html
         return rendering
 
-    def __init__(self, pages, metadata, url_fetcher, font_config,
-                 optimize_size):
+    def __init__(self, pages, metadata, url_fetcher, font_config):
         #: A list of :class:`Page` objects.
         self.pages = pages
         #: A :class:`DocumentMetadata` object.
@@ -246,9 +275,6 @@ class Document:
         # rendering is destroyed. This is needed as font_config.__del__ removes
         # fonts that may be used when rendering
         self.font_config = font_config
-        # Set of flags for PDF size optimization. Can contain "images" and
-        # "fonts".
-        self._optimize_size = optimize_size
 
     def build_element_structure(self, structure, etree_element=None):
         if etree_element is None:
@@ -288,8 +314,7 @@ class Document:
         elif not isinstance(pages, list):
             pages = list(pages)
         return type(self)(
-            pages, self.metadata, self.url_fetcher, self.font_config,
-            self._optimize_size)
+            pages, self.metadata, self.url_fetcher, self.font_config)
 
     def make_bookmark_tree(self, scale=1, transform_pages=False):
         """Make a tree of all bookmarks in the document.
@@ -324,9 +349,7 @@ class Document:
                 page_number, matrix)
         return root
 
-    def write_pdf(self, target=None, zoom=1, attachments=None, finisher=None,
-                  identifier=None, variant=None, version=None,
-                  custom_metadata=False):
+    def write_pdf(self, target=None, zoom=1, finisher=None, **options):
         """Paint the pages in a PDF file, with metadata.
 
         :type target:
@@ -339,40 +362,48 @@ class Document:
             All CSS units are affected, including physical units like
             ``cm`` and named sizes like ``A4``.  For values other than
             1, the physical CSS units will thus be "wrong".
-        :param list attachments: A list of additional file attachments for the
-            generated PDF document or :obj:`None`. The list's elements are
-            :class:`weasyprint.Attachment` objects, filenames, URLs or
-            file-like objects.
-        :param finisher: A finisher function, that accepts the document and a
-            :class:`pydyf.PDF` object as parameters, can be passed to perform
+        :type finisher: :term:`callable`
+        :param finisher:
+            A finisher function or callable that accepts the document and a
+            :class:`pydyf.PDF` object as parameters. Can be passed to perform
             post-processing on the PDF right before the trailer is written.
-        :param bytes identifier: A bytestring used as PDF file identifier.
-        :param str variant: A PDF variant name.
-        :param str version: A PDF version number.
-        :param bool custom_metadata: A boolean defining whether custom HTML
-            metadata should be stored in the generated PDF.
+        :param options:
+            The ``options`` parameter includes by default the
+            :data:`weasyprint.DEFAULT_OPTIONS` values.
         :returns:
             The PDF as :obj:`bytes` if ``target`` is not provided or
             :obj:`None`, otherwise :obj:`None` (the PDF is written to
             ``target``).
 
         """
-        pdf = generate_pdf(
-            self, target, zoom, attachments, self._optimize_size, identifier,
-            variant, version, custom_metadata)
+        new_options = DEFAULT_OPTIONS.copy()
+        new_options.update(options)
+        options = new_options
+
+        # Set default PDF version for PDF variants.
+        if variant := options['pdf_variant']:
+            _, properties = VARIANTS[variant]
+            if 'version' in properties and not options['pdf_version']:
+                options['pdf_version'] = properties['version']
+            if 'identifier' in properties and not options['pdf_identifier']:
+                options['pdf_identifier'] = properties['identifier']
+
+        pdf = generate_pdf(self, target, zoom, **options)
 
         if finisher:
             finisher(self, pdf)
 
-        output = io.BytesIO()
-        pdf.write(output, version=pdf.version, identifier=identifier)
+        identifier = options['pdf_identifier']
+        compress = not options['uncompressed_pdf']
+        version = options['pdf_version']
 
         if target is None:
+            output = io.BytesIO()
+            pdf.write(output, version, identifier, compress)
             return output.getvalue()
+
+        if hasattr(target, 'write'):
+            pdf.write(target, version, identifier, compress)
         else:
-            output.seek(0)
-            if hasattr(target, 'write'):
-                shutil.copyfileobj(output, target)
-            else:
-                with open(target, 'wb') as fd:
-                    shutil.copyfileobj(output, fd)
+            with open(target, 'wb') as fd:
+                pdf.write(fd, version, identifier, compress)
