@@ -1,14 +1,22 @@
 """Fetch and decode images in various formats."""
 
+import io
 import math
+import struct
 from hashlib import md5
 from io import BytesIO
 from itertools import cycle
 from math import inf
+from pathlib import Path
+from urllib.parse import urlparse
+from urllib.request import url2pathname
 from xml.etree import ElementTree
 
+import pydyf
 from PIL import Image, ImageFile, ImageOps
+from tinycss2.color4 import parse_color
 
+from . import DEFAULT_OPTIONS
 from .layout.percent import percentage
 from .logger import LOGGER
 from .svg import SVG
@@ -33,32 +41,220 @@ class ImageLoadingError(ValueError):
 
 
 class RasterImage:
-    def __init__(self, pillow_image, image_id, optimize_size):
-        pillow_image.id = image_id
-        self._pillow_image = pillow_image
-        self._optimize_size = optimize_size
-        self._intrinsic_width = pillow_image.width
-        self._intrinsic_height = pillow_image.height
-        self._intrinsic_ratio = (
-            self._intrinsic_width / self._intrinsic_height
-            if self._intrinsic_height != 0 else inf)
+    def __init__(self, pillow_image, image_id, image_data, filename=None,
+                 cache=None, orientation='none', options=DEFAULT_OPTIONS):
+        # Transpose image
+        original_pillow_image = pillow_image
+        pillow_image = rotate_pillow_image(pillow_image, orientation)
+        if original_pillow_image is not pillow_image:
+            # Keep image format as it is discarded by transposition
+            pillow_image.format = original_pillow_image.format
+            # Discard original data, as the image has been transformed
+            image_data = filename = None
 
-    def get_intrinsic_size(self, image_resolution, font_size):
-        return (
-            self._intrinsic_width / image_resolution,
-            self._intrinsic_height / image_resolution,
-            self._intrinsic_ratio)
+        self.id = image_id
+        self._cache = {} if cache is None else cache
+        self._jpeg_quality = jpeg_quality = options['jpeg_quality']
+        self._dpi = options['dpi']
+
+        if 'transparency' in pillow_image.info:
+            pillow_image = pillow_image.convert('RGBA')
+        elif pillow_image.mode in ('1', 'P', 'I'):
+            pillow_image = pillow_image.convert('RGB')
+
+        self.mode = pillow_image.mode
+        self.width = pillow_image.width
+        self.height = pillow_image.height
+        self.ratio = (self.width / self.height) if self.height != 0 else inf
+        self.optimize = optimize = options['optimize_images']
+
+        # The presence of the APP14 segment indicates an Adobe image with
+        # inverted CMYK data. Specify a Decode Array to invert it again back to
+        # normal. See https://github.com/Kozea/WeasyPrint/pull/2179.
+        app14 = getattr(original_pillow_image, 'app', {}).get('APP14')
+        self.invert_colors = self.mode == 'CMYK' and app14 is not None
+
+        if pillow_image.format in ('JPEG', 'MPO'):
+            self.format = 'JPEG'
+            if image_data is None or optimize or jpeg_quality is not None:
+                image_file = io.BytesIO()
+                options = {'format': 'JPEG', 'optimize': optimize}
+                if self._jpeg_quality is not None:
+                    options['quality'] = self._jpeg_quality
+                pillow_image.save(image_file, **options)
+                image_data = image_file.getvalue()
+                filename = None
+        else:
+            self.format = 'PNG'
+            if image_data is None or optimize or pillow_image.format != 'PNG':
+                image_file = io.BytesIO()
+                pillow_image.save(image_file, format='PNG', optimize=optimize)
+                image_data = image_file.getvalue()
+                filename = None
+        self.image_data = self.cache_image_data(image_data, filename)
+
+    def get_intrinsic_size(self, resolution, font_size):
+        return self.width / resolution, self.height / resolution, self.ratio
 
     def draw(self, stream, concrete_width, concrete_height, image_rendering):
-        if self._intrinsic_width <= 0 or self._intrinsic_height <= 0:
+        if self.width <= 0 or self.height <= 0:
             return
 
-        image_name = stream.add_image(
-            self._pillow_image, image_rendering, self._optimize_size)
+        interpolate = image_rendering == 'auto'
+        ratio = 1
+        if self._dpi:
+            pt_to_in = 4 / 3 / 96
+            width_inches = abs(concrete_width * stream.ctm[0][0] * pt_to_in)
+            height_inches = abs(concrete_height * stream.ctm[1][1] * pt_to_in)
+            dpi = max(self.width / width_inches, self.height / height_inches)
+            if dpi > self._dpi:
+                ratio = self._dpi / dpi
+        image_name = stream.add_image(self, interpolate, ratio)
+
         stream.transform(
             concrete_width, 0, 0, -concrete_height, 0, concrete_height)
         stream.draw_x_object(image_name)
 
+    def cache_image_data(self, data, filename=None, slot='source'):
+        if filename:
+            return LazyLocalImage(filename)
+        else:
+            key = f'{self.id}-{slot}-{self._dpi or ""}'
+            return LazyImage(self._cache, key, data)
+
+    def get_x_object(self, interpolate, dpi_ratio):
+        if dpi_ratio == 1:
+            width, height = self.width, self.height
+        else:
+            thumbnail = Image.open(io.BytesIO(self.image_data.data))
+            width = max(1, int(round(self.width * dpi_ratio)))
+            height = max(1, int(round(self.height * dpi_ratio)))
+            thumbnail.thumbnail((width, height))
+            image_file = io.BytesIO()
+            thumbnail.save(
+                image_file, format=thumbnail.format, optimize=self.optimize)
+            width, height = thumbnail.width, thumbnail.height
+            self.image_data = self.cache_image_data(image_file.getvalue())
+
+        if self.mode in ('RGB', 'RGBA'):
+            color_space = '/DeviceRGB'
+        elif self.mode in ('L', 'LA'):
+            color_space = '/DeviceGray'
+        elif self.mode == 'CMYK':
+            color_space = '/DeviceCMYK'
+        else:
+            LOGGER.warning('Unknown image mode: %s', self.mode)
+            color_space = '/DeviceRGB'
+
+        extra = pydyf.Dictionary({
+            'Type': '/XObject',
+            'Subtype': '/Image',
+            'Width': width,
+            'Height': height,
+            'ColorSpace': color_space,
+            'BitsPerComponent': 8,
+            'Interpolate': 'true' if interpolate else 'false',
+        })
+
+        if self.format == 'JPEG':
+            if self.invert_colors:
+                extra['Decode'] = pydyf.Array((1, 0) * 4)
+            extra['Filter'] = '/DCTDecode'
+            return pydyf.Stream([self.image_data], extra)
+
+        extra['Filter'] = '/FlateDecode'
+        extra['DecodeParms'] = pydyf.Dictionary({
+            # Predictor 15 specifies that we're providing PNG data,
+            # ostensibly using an "optimum predictor", but doesn't actually
+            # matter as long as the predictor value is 10+ according to the
+            # spec. (Other PNG predictor values assert that we're using
+            # specific predictors that we don't want to commit to, but
+            # "optimum" can vary.)
+            'Predictor': 15,
+            'Columns': width,
+        })
+        if self.mode in ('RGB', 'RGBA'):
+            # Defaults to 1.
+            extra['DecodeParms']['Colors'] = 3
+        if self.mode in ('RGBA', 'LA'):
+            # Remove alpha channel from image
+            pillow_image = Image.open(io.BytesIO(self.image_data.data))
+            alpha = pillow_image.getchannel('A')
+            pillow_image = pillow_image.convert(self.mode[:-1])
+            png_data = self._get_png_data(pillow_image)
+            # Save alpha channel as mask
+            alpha_data = self._get_png_data(alpha)
+            stream = self.cache_image_data(alpha_data, slot='streamalpha')
+            extra['SMask'] = pydyf.Stream([stream], extra={
+                'Filter': '/FlateDecode',
+                'Type': '/XObject',
+                'Subtype': '/Image',
+                'DecodeParms': pydyf.Dictionary({
+                    'Predictor': 15,
+                    'Columns': width,
+                }),
+                'Width': width,
+                'Height': height,
+                'ColorSpace': '/DeviceGray',
+                'BitsPerComponent': 8,
+                'Interpolate': 'true' if interpolate else 'false',
+            })
+        else:
+            png_data = self._get_png_data(
+                Image.open(io.BytesIO(self.image_data.data)))
+
+        return pydyf.Stream([self.cache_image_data(png_data, slot='stream')], extra)
+
+    @staticmethod
+    def _get_png_data(pillow_image):
+        image_file = BytesIO()
+        pillow_image.save(image_file, format='PNG')
+
+        # Read the PNG header, then discard it because we know it's a PNG. If
+        # this weren't just output from Pillow, we should actually check it.
+        image_file.seek(8)
+
+        png_data = []
+        raw_chunk_length = image_file.read(4)
+        # PNG files consist of a series of chunks.
+        while raw_chunk_length:
+            # Each chunk begins with its data length (four bytes, may be zero),
+            # then its type (four ASCII characters), then the data, then four
+            # bytes of a CRC.
+            chunk_length, = struct.unpack('!I', raw_chunk_length)
+            chunk_type = image_file.read(4)
+            if chunk_type == b'IDAT':
+                png_data.append(image_file.read(chunk_length))
+            else:
+                image_file.seek(chunk_length, io.SEEK_CUR)
+            # We aren't checking the CRC, we assume this is a valid PNG.
+            image_file.seek(4, io.SEEK_CUR)
+            raw_chunk_length = image_file.read(4)
+
+        return b''.join(png_data)
+
+
+class LazyImage(pydyf.Object):
+    def __init__(self, cache, key, data):
+        super().__init__()
+        self._key = key
+        self._cache = cache
+        cache[key] = data
+
+    @property
+    def data(self):
+        return self._cache[self._key]
+
+
+class LazyLocalImage(pydyf.Object):
+    def __init__(self, filename):
+        super().__init__()
+        self._filename = filename
+
+    @property
+    def data(self):
+        return Path(self._filename).read_bytes()
+
 
 class SVGImage:
     def __init__(self, tree, base_url, url_fetcher, context):
@@ -86,80 +282,98 @@ class SVGImage:
         return width, height, ratio
 
     def draw(self, stream, concrete_width, concrete_height, image_rendering):
-        self._svg.draw(
-            stream, concrete_width, concrete_height, self._base_url,
-            self._url_fetcher, self._context)
+        try:
+            self._svg.draw(
+                stream, concrete_width, concrete_height, self._base_url,
+                self._url_fetcher, self._context)
+        except BaseException as exception:
+            LOGGER.error('Failed to render SVG image %s', self._base_url)
+            LOGGER.debug('Error while rendering SVG image:', exc_info=exception)
 
 
-def get_image_from_uri(cache, url_fetcher, optimize_size, url,
-                       forced_mime_type=None, context=None,
-                       orientation='from-image'):
+def get_image_from_uri(cache, url_fetcher, options, url, forced_mime_type=None,
+                       context=None, orientation='from-image'):
     """Get an Image instance from an image URI."""
     if url in cache:
         return cache[url]
 
     try:
         with fetch(url_fetcher, url) as result:
+            parsed_url = urlparse(result.get('redirected_url'))
+            if parsed_url.scheme == 'file':
+                filename = url2pathname(parsed_url.path)
+            else:
+                filename = None
             if 'string' in result:
                 string = result['string']
             else:
                 string = result['file_obj'].read()
             mime_type = forced_mime_type or result['mime_type']
 
-            image = None
-            svg_exceptions = []
-            # Try to rely on given mimetype for SVG
-            if mime_type == 'image/svg+xml':
+        image = None
+        svg_exceptions = []
+        # Try to rely on given mimetype for SVG
+        if mime_type == 'image/svg+xml':
+            try:
+                tree = ElementTree.fromstring(string)
+                image = SVGImage(tree, url, url_fetcher, context)
+            except Exception as svg_exception:
+                svg_exceptions.append(svg_exception)
+        # Try pillow for raster images, or for failing SVG
+        if image is None:
+            try:
+                pillow_image = Image.open(BytesIO(string))
+            except Exception as raster_exception:
+                if mime_type == 'image/svg+xml':
+                    # Tried SVGImage then Pillow for a SVG, abort
+                    raise ImageLoadingError.from_exception(svg_exceptions[0])
                 try:
+                    # Last chance, try SVG
                     tree = ElementTree.fromstring(string)
                     image = SVGImage(tree, url, url_fetcher, context)
-                except Exception as svg_exception:
-                    svg_exceptions.append(svg_exception)
-            # Try pillow for raster images, or for failing SVG
-            if image is None:
-                try:
-                    pillow_image = Image.open(BytesIO(string))
-                except Exception as raster_exception:
-                    if mime_type == 'image/svg+xml':
-                        # Tried SVGImage then Pillow for a SVG, abort
-                        raise ImageLoadingError.from_exception(
-                            svg_exceptions[0])
-                    try:
-                        # Last chance, try SVG
-                        tree = ElementTree.fromstring(string)
-                        image = SVGImage(tree, url, url_fetcher, context)
-                    except Exception:
-                        # Tried Pillow then SVGImage for a raster, abort
-                        raise ImageLoadingError.from_exception(
-                            raster_exception)
-                else:
-                    # Store image id to enable cache in Stream.add_image
-                    image_id = md5(url.encode()).hexdigest()
-                    # Keep image format as it is discarded by transposition
-                    image_format = pillow_image.format
-                    if orientation == 'from-image':
-                        if 'exif' in pillow_image.info:
-                            pillow_image = ImageOps.exif_transpose(
-                                pillow_image)
-                    elif orientation != 'none':
-                        angle, flip = orientation
-                        if angle > 0:
-                            rotation = getattr(
-                                Image.Transpose, f'ROTATE_{angle}')
-                            pillow_image = pillow_image.transpose(rotation)
-                        if flip:
-                            pillow_image = pillow_image.transpose(
-                                Image.Transpose.FLIP_LEFT_RIGHT)
-                    pillow_image.format = image_format
-                    image = RasterImage(pillow_image, image_id, optimize_size)
+                except Exception:
+                    # Tried Pillow then SVGImage for a raster, abort
+                    raise ImageLoadingError.from_exception(raster_exception)
+            else:
+                # Store image id to enable cache in Stream.add_image
+                image_id = md5(url.encode(), usedforsecurity=False).hexdigest()
+                image = RasterImage(
+                    pillow_image, image_id, string, filename, cache,
+                    orientation, options)
 
     except (URLFetchingError, ImageLoadingError) as exception:
         LOGGER.error('Failed to load image at %r: %s', url, exception)
+        LOGGER.debug('Error while loading image:', exc_info=exception)
         image = None
+
     cache[url] = image
     return image
 
 
+def rotate_pillow_image(pillow_image, orientation):
+    """Return a copy of a Pillow image with modified orientation.
+
+    If orientation is not changed, return the same image.
+
+    """
+    image_format = pillow_image.format
+    if orientation == 'from-image':
+        if 'exif' in pillow_image.info:
+            pillow_image = ImageOps.exif_transpose(pillow_image)
+    elif orientation != 'none':
+        angle, flip = orientation
+        if angle > 0:
+            rotation = getattr(Image.Transpose, f'ROTATE_{angle}')
+            pillow_image = pillow_image.transpose(rotation)
+        if flip:
+            pillow_image = pillow_image.transpose(
+                Image.Transpose.FLIP_LEFT_RIGHT)
+
+    # Keep image format as it is discarded by transposition
+    pillow_image.format = image_format
+    return pillow_image
+
+
 def process_color_stops(vector_length, positions):
     """Give color stops positions on the gradient vector.
 
@@ -228,6 +442,7 @@ def gradient_average_color(colors, positions):
     """
     https://drafts.csswg.org/css-images-3/#gradient-average-color
     """
+    # TODO: handle color spaces.
     nb_stops = len(positions)
     assert nb_stops > 1
     assert nb_stops == len(colors)
@@ -248,9 +463,13 @@ def gradient_average_color(colors, positions):
             result_g += premul_g[j] * weight
             result_b += premul_b[j] * weight
             result_a += alpha[j] * weight
-    # Un-premultiply:
-    return (result_r / result_a, result_g / result_a,
-            result_b / result_a, result_a) if result_a != 0 else (0, 0, 0, 0)
+    # Un-premultiply.
+    if result_a == 0:
+        return parse_color('transparent')
+    else:
+        return parse_color(
+            f'rgb({result_r / result_a * 255} {result_g / result_a * 255} '
+            f'{result_b / result_a * 255}/{ result_a })')
 
 
 class Gradient:
@@ -267,16 +486,12 @@ class Gradient:
         return None, None, None
 
     def draw(self, stream, concrete_width, concrete_height, _image_rendering):
-        # TODO: handle color spaces
         scale_y, type_, points, positions, colors = self.layout(
             concrete_width, concrete_height)
 
         if type_ == 'solid':
             stream.rectangle(0, 0, concrete_width, concrete_height)
-            red, green, blue, alpha = colors[0]
-            stream.set_color_rgb(red, green, blue)
-            if alpha != 1:
-                stream.set_alpha(alpha, stroke=False)
+            stream.set_color(colors[0])
             stream.fill()
             return
 
@@ -284,8 +499,9 @@ class Gradient:
         alpha_couples = [
             (alphas[i], alphas[i + 1])
             for i in range(len(alphas) - 1)]
+        # TODO: handle other color spaces.
         color_couples = [
-            [colors[i][:3], colors[i + 1][:3], 1]
+            [colors[i].to('srgb')[:3], colors[i + 1].to('srgb')[:3], 1]
             for i in range(len(colors) - 1)]
 
         # Premultiply colors
@@ -309,6 +525,7 @@ class Gradient:
             for c0, c1, n in color_couples)
         function = stream.create_stitching_function(
             domain, encode, bounds, sub_functions)
+        # TODO: handle other color spaces.
         shading = stream.add_shading(
             shading_type, 'RGB', domain, points, extend, function)
         stream.transform(d=scale_y)
@@ -328,7 +545,7 @@ class Gradient:
             alpha_stream.transform(d=scale_y)
             alpha_stream.stream = [f'/{alpha_shading.id} sh']
 
-        stream.shading(shading.id)
+        stream.paint_shading(shading.id)
 
     def layout(self, width, height):
         """Get layout information about the gradient.
@@ -416,9 +633,9 @@ class LinearGradient(Gradient):
                 for i in range(len(positions) - 1)]
 
             # Create cycles used to add colors
-            next_steps = cycle([0] + position_steps)
+            next_steps = cycle((0, *position_steps))
             next_colors = cycle(colors)
-            previous_steps = cycle([0] + position_steps[::-1])
+            previous_steps = cycle((0, *position_steps[::-1]))
             previous_colors = cycle(colors[::-1])
 
             # Add colors after last step
@@ -617,8 +834,7 @@ class RadialGradient(Gradient):
                 new_positions = [
                     position - 1 - full_repeat for position
                     in original_positions[-(i - 1):]]
-                positions = (
-                    [ratio - 1 - full_repeat] + new_positions + positions)
+                positions = (ratio - 1 - full_repeat, *new_positions, *positions)
                 return points, positions, colors
 
     def _resolve_size(self, width, height, center_x, center_y):
