@@ -7,6 +7,7 @@ from .absolute import AbsolutePlaceholder, absolute_layout
 from .column import columns_layout
 from .flex import flex_layout
 from .float import avoid_collisions, float_layout, get_clearance
+from .grid import grid_layout
 from .inline import iter_line_boxes
 from .min_max import handle_min_max_width
 from .percent import resolve_percentages, resolve_position_percentages
@@ -48,8 +49,7 @@ def block_level_layout(context, box, bottom_space, skip_stack,
                     if not context.forced_break:
                         box.margin_top = 0
 
-        collapsed_margin = collapse_margin(
-            adjoining_margins + [box.margin_top])
+        collapsed_margin = collapse_margin([*adjoining_margins, box.margin_top])
         box.clearance = get_clearance(context, box, collapsed_margin)
         if box.clearance is not None:
             top_border_edge = box.position_y + collapsed_margin + box.clearance
@@ -82,9 +82,13 @@ def block_level_layout_switch(context, box, bottom_space, skip_stack,
         result = flex_layout(
             context, box, bottom_space, skip_stack, containing_block,
             page_is_empty, absolute_boxes, fixed_boxes)
+    elif isinstance(box, boxes.GridBox):
+        result = grid_layout(
+            context, box, bottom_space, skip_stack, containing_block,
+            page_is_empty, absolute_boxes, fixed_boxes)
     else:  # pragma: no cover
         raise TypeError(f'Layout for {type(box).__name__} not handled yet')
-    return result + (None,)
+    return (*result, None)
 
 
 def block_box_layout(context, box, bottom_space, skip_stack,
@@ -111,7 +115,7 @@ def block_box_layout(context, box, bottom_space, skip_stack,
                     context, box, bottom_space, skip_stack,
                     containing_block, page_is_empty, absolute_boxes,
                     fixed_boxes, adjoining_margins)
-        return result + (None,)
+        return (*result, None)
     elif box.is_table_wrapper:
         table_wrapper_width(
             context, box, (containing_block.width, containing_block.height))
@@ -120,8 +124,11 @@ def block_box_layout(context, box, bottom_space, skip_stack,
     result = block_container_layout(
         context, box, bottom_space, skip_stack, page_is_empty,
         absolute_boxes, fixed_boxes, adjoining_margins, discard, max_lines)
-    new_box = result[0]
-    if new_box and new_box.is_table_wrapper:
+    # TODO: columns and flex items shouldn't be block boxes, this condition
+    # would then be useless when this is fixed.
+    if not (new_box := result[0]) or new_box.is_column or new_box.is_flex_item:
+        return result
+    if new_box.is_table_wrapper or new_box.establishes_formatting_context():
         # Don't collide with floats
         # https://www.w3.org/TR/CSS21/visuren.html#floats
         position_x, position_y, _ = avoid_collisions(
@@ -228,12 +235,17 @@ def relative_positioning(box, containing_block):
 def _out_of_flow_layout(context, box, index, child, new_children,
                         page_is_empty, absolute_boxes, fixed_boxes,
                         adjoining_margins, bottom_space):
-    stop = False
-    resume_at = None
-    new_child = None
-    out_of_flow_resume_at = None
+    stop = False  # whether we should stop parent rendering after this layout
+    resume_at = None  # where to resume in-flow rendering
+    new_child = None  # child rendered by this layout
+    out_of_flow_resume_at = None  # where to resume out-of-flow rendering
 
+    # Add the parent’s collapsing margins to shift the child’s position. Don’t
+    # include the out-of-flow child’s top margin because it doesn’t collapse
+    # with its parent.
     child.position_y += collapse_margin(adjoining_margins)
+
+    # Absolute child layout: create placeholder.
     if child.is_absolutely_positioned():
         new_child = placeholder = AbsolutePlaceholder(child)
         placeholder.index = index
@@ -242,26 +254,42 @@ def _out_of_flow_layout(context, box, index, child, new_children,
             absolute_boxes.append(placeholder)
         else:
             fixed_boxes.append(placeholder)
+
+    # Float child layout.
     elif child.is_floated():
         new_child, out_of_flow_resume_at = float_layout(
             context, child, box, absolute_boxes, fixed_boxes, bottom_space,
             skip_stack=None)
-        # New page if overflow
+
+        # Check that child doesn’t overflow page.
         page_overflow = context.overflows_page(
             bottom_space, new_child.position_y + new_child.height)
-        if (page_is_empty and not new_children) or not page_overflow:
+        add_child = (
+            (page_is_empty and not new_children) or
+            not page_overflow or
+            box.is_monolithic())
+        if add_child:
+            # Child fits or has to fit, add it.
             new_child.index = index
             new_children.append(new_child)
         else:
+            # Child doesn’t fit and we can break, find where to break and stop
+            # parent rendering.
             last_in_flow_child = find_last_in_flow_child(new_children)
             page_break = block_level_page_break(last_in_flow_child, child)
             resume_at = {index: None}
+            out_of_flow_resume_at = None
+            stop = True
             if new_children and avoid_page_break(page_break, context):
+                # Can’t break inside float, find an earlier page break.
                 result = find_earlier_page_break(
                     context, new_children, absolute_boxes, fixed_boxes)
                 if result:
-                    new_children, resume_at = result
-            stop = True
+                    # Earlier page break found, drop whole child rendering.
+                    new_children[:], resume_at = result
+                    new_child = None
+
+    # Running element layout.
     elif child.is_running():
         running_name = child.style['position'][1]
         page = context.current_page
@@ -277,6 +305,7 @@ def _break_line(context, box, line, new_children, lines_iterator,
     if over_orphans < 0 and not page_is_empty:
         # Reached the bottom of the page before we had
         # enough lines for orphans, cancel the whole box.
+        remove_placeholders(context, line.children, absolute_boxes, fixed_boxes)
         return True, False, resume_at
     # How many lines we need on the next page to satisfy widows
     # -1 for the current line.
@@ -288,6 +317,7 @@ def _break_line(context, box, line, new_children, lines_iterator,
                 break
     if needed > over_orphans and not page_is_empty:
         # Total number of lines < orphans + widows
+        remove_placeholders(context, line.children, absolute_boxes, fixed_boxes)
         return True, False, resume_at
     if needed and needed <= over_orphans:
         # Remove lines to keep them for the next page
@@ -449,7 +479,7 @@ def _in_flow_layout(context, box, index, child, new_children, page_is_empty,
                     if not context.forced_break:
                         child_margin_top = 0
             new_collapsed_margin = collapse_margin(
-                adjoining_margins + [child_margin_top])
+                [*adjoining_margins, child_margin_top])
             collapsed_margin_difference = (
                 new_collapsed_margin - old_collapsed_margin)
             for previous_new_child in new_children:
@@ -499,16 +529,19 @@ def _in_flow_layout(context, box, index, child, new_children, page_is_empty,
                 new_child.content_box_y() + new_child.height)
             new_position_y = (
                 new_child.border_box_y() + new_child.border_height())
-            page_overflow = context.overflows_page(
+            content_page_overflow = context.overflows_page(
                 bottom_space, new_content_position_y)
-            if page_overflow and not page_is_empty_with_no_children:
+            border_page_overflow = context.overflows_page(
+                bottom_space, new_position_y)
+            can_break = not (
+                page_is_empty_with_no_children or box.is_monolithic())
+            if can_break and content_page_overflow:
                 # The child content overflows the page area, display it on the
                 # next page.
                 remove_placeholders(
                     context, [new_child], absolute_boxes, fixed_boxes)
                 new_child = None
-            elif not page_is_empty_with_no_children and context.overflows_page(
-                    bottom_space, new_position_y):
+            elif can_break and border_page_overflow:
                 # The child border/padding overflows the page area, do the
                 # layout again with a higher bottom_space value.
                 remove_placeholders(
@@ -600,7 +633,7 @@ def block_container_layout(context, box, bottom_space, skip_stack,
     # block_container_layout, there's probably a better solution.
     assert isinstance(box, (boxes.BlockContainerBox, boxes.FlexBox))
 
-    if establishes_formatting_context(box):
+    if box.establishes_formatting_context():
         context.create_block_formatting_context()
 
     is_start = skip_stack is None
@@ -622,7 +655,8 @@ def block_container_layout(context, box, bottom_space, skip_stack,
 
     collapsing_with_children = not (
         box.border_top_width or box.padding_top or box.is_flex_item or
-        establishes_formatting_context(box) or box.is_for_root_element)
+        box.is_grid_item or box.establishes_formatting_context() or
+        box.is_for_root_element)
     if collapsing_with_children:
         # Not counting margins in adjoining_margins, if any
         # (there are not padding or borders, see above)
@@ -718,6 +752,11 @@ def block_container_layout(context, box, bottom_space, skip_stack,
                 None, None, {'break': 'any', 'page': page}, [], False,
                 max_lines)
         elif stop:
+            if box.height != 'auto':
+                if context.overflows(box.position_y + box.height, position_y):
+                    # Box heigh is fixed and it doesn’t overflow page, forget
+                    # overflowing children.
+                    resume_at = None
             adjoining_margins = []
             break
 
@@ -731,10 +770,9 @@ def block_container_layout(context, box, bottom_space, skip_stack,
     if (box_is_fragmented and
             avoid_page_break(box.style['break_inside'], context) and
             not page_is_empty):
-        for footnote in all_footnotes:
-            context.unlayout_footnote(footnote)
-        return (
-            None, None, {'break': 'any', 'page': None}, [], False, max_lines)
+        remove_placeholders(
+            context, [*new_children, *box.children[skip:]], absolute_boxes, fixed_boxes)
+        return None, None, {'break': 'any', 'page': None}, [], False, max_lines
 
     for key, value in broken_out_of_flow.items():
         context.broken_out_of_flow[key] = value
@@ -765,7 +803,7 @@ def block_container_layout(context, box, bottom_space, skip_stack,
 
     if (box.border_bottom_width or
             box.padding_bottom or
-            establishes_formatting_context(box) or
+            box.establishes_formatting_context() or
             box.is_for_root_element or
             box.is_table_wrapper):
         position_y += collapse_margin(adjoining_margins)
@@ -802,7 +840,7 @@ def block_container_layout(context, box, bottom_space, skip_stack,
     for child in new_box.children:
         relative_positioning(child, (new_box.width, new_box.height))
 
-    if establishes_formatting_context(new_box):
+    if new_box.establishes_formatting_context():
         context.finish_block_formatting_context(new_box)
 
     if discard or not box_is_fragmented:
@@ -840,30 +878,6 @@ def collapse_margin(adjoining_margins):
     return max(positives) + min(negatives)
 
 
-def establishes_formatting_context(box):
-    """Return whether a box establishes a block formatting context.
-
-    See https://www.w3.org/TR/CSS2/visuren.html#block-formatting
-
-    """
-    return (
-        box.is_floated()
-    ) or (
-        box.is_absolutely_positioned()
-    ) or (
-        # TODO: columns shouldn't be block boxes, this condition would then be
-        # useless when this is fixed
-        box.is_column
-    ) or (
-        isinstance(box, boxes.BlockContainerBox) and
-        not isinstance(box, boxes.BlockBox)
-    ) or (
-        isinstance(box, boxes.BlockBox) and box.style['overflow'] != 'visible'
-    ) or (
-        'flow-root' in box.style['display']
-    )
-
-
 def block_level_page_break(sibling_before, sibling_after):
     """Get the correct page break value between siblings.
 
@@ -889,7 +903,7 @@ def block_level_page_break(sibling_before, sibling_after):
     box = sibling_before
     while isinstance(box, block_parallel_box_types):
         values.append(box.style['break_after'])
-        if not (isinstance(box, boxes.ParentBox) and box.children):
+        if not box.children:
             break
         box = box.children[-1]
     values.reverse()  # Have them in tree order
@@ -897,7 +911,7 @@ def block_level_page_break(sibling_before, sibling_after):
     box = sibling_after
     while isinstance(box, block_parallel_box_types):
         values.append(box.style['break_before'])
-        if not (isinstance(box, boxes.ParentBox) and box.children):
+        if not box.children:
             break
         box = box.children[0]
 
@@ -985,7 +999,7 @@ def find_earlier_page_break(context, children, absolute_boxes, fixed_boxes):
                 if result:
                     new_grand_children, resume_at = result
                     new_child = child.copy_with_children(new_grand_children)
-                    new_children = list(children[:index]) + [new_child]
+                    new_children = [*children[:index], new_child]
 
                     # Re-add footer at the end of split table
                     # TODO: fix table height and footer position
