@@ -5,16 +5,16 @@ from math import inf
 
 from ..css import computed_from_cascaded
 from ..css.computed_values import character_ratio, strut_layout
-from ..formatting_structure import boxes
+from ..formatting_structure import boxes, build
 from ..text.line_break import can_break_text, create_layout, split_first_line
 from .absolute import AbsolutePlaceholder, absolute_layout
 from .flex import flex_layout
 from .float import avoid_collisions, float_layout
+from .grid import grid_layout
 from .leader import handle_leader
 from .min_max import handle_min_max_width
 from .percent import resolve_one_percentage, resolve_percentages
-from .preferred import (
-    inline_min_content_width, shrink_to_fit, trailing_whitespace_size)
+from .preferred import inline_min_content_width, shrink_to_fit, trailing_whitespace_size
 from .replaced import inline_replaced_box_layout
 from .table import find_in_flow_baseline, table_wrapper_width
 
@@ -192,13 +192,15 @@ def skip_first_whitespace(box, skip_stack):
     if isinstance(box, boxes.TextBox):
         assert next_skip_stack is None
         white_space = box.style['white_space']
-        length = len(box.text)
-        if index == length:
+        text = box.text.encode()
+        if index == len(text):
             # Starting a the end of the TextBox, no text to see: Continue
             return 'continue'
         if white_space in ('normal', 'nowrap', 'pre-line'):
-            while index < length and box.text[index] == ' ':
+            text = text[index:]
+            while text and text.startswith(b' '):
                 index += 1
+                text = text[1:]
         return {index: None} if index else None
 
     if isinstance(box, (boxes.LineBox, boxes.InlineBox)):
@@ -292,8 +294,7 @@ def first_letter_to_box(box, skip_stack, first_letter_style):
                 letter_box = boxes.InlineBox(
                     f'{box.element_tag}::first-letter', letter_style,
                     box.element, [child])
-                box.children = (
-                    (letter_box,) + tuple(box.children[1:]))
+                box.children = ((letter_box, *box.children[1:]))
             elif child.text:
                 character_found = False
                 if skip_stack:
@@ -323,7 +324,7 @@ def first_letter_to_box(box, skip_stack, first_letter_style):
                             f'{box.element_tag}::first-letter', letter_style,
                             box.element, first_letter)
                         letter_box.children = (text_box,)
-                        box.children = (letter_box,) + tuple(box.children)
+                        box.children = (letter_box, *box.children)
                     else:
                         letter_box = boxes.BlockBox(
                             f'{box.element_tag}::first-letter',
@@ -337,7 +338,8 @@ def first_letter_to_box(box, skip_stack, first_letter_style):
                             f'{box.element_tag}::first-letter', letter_style,
                             box.element, first_letter)
                         line_box.children = (text_box,)
-                        box.children = (letter_box,) + tuple(box.children)
+                        box.children = (letter_box, *box.children)
+                    build.process_text_transform(text_box)
                     if skip_stack and child_skip_stack:
                         index, = skip_stack
                         (child_index, grandchild_skip_stack), = (
@@ -479,7 +481,7 @@ def split_inline_level(context, box, position_x, max_x, bottom_space,
             if skip is None:
                 last_letter = box.text[-1]
             else:
-                last_letter = box.text[skip - 1]
+                last_letter = box.text.encode()[:skip].decode()[-1]
         else:
             first_letter = last_letter = None
     elif isinstance(box, boxes.InlineBox):
@@ -515,6 +517,18 @@ def split_inline_level(context, box, position_x, max_x, bottom_space,
         preserved_line_break = False
         first_letter = '\u2e80'
         last_letter = '\u2e80'
+    elif isinstance(box, boxes.InlineGridBox):
+        box.position_x = position_x
+        box.position_y = 0
+        for side in ('top', 'right', 'bottom', 'left'):
+            if getattr(box, f'margin_{side}') == 'auto':
+                setattr(box, f'margin_{side}', 0)
+        new_box, resume_at, _, _, _ = grid_layout(
+            context, box, -inf, skip_stack, containing_block, False,
+            absolute_boxes, fixed_boxes)
+        preserved_line_break = False
+        first_letter = '\u2e80'
+        last_letter = '\u2e80'
     else:  # pragma: no cover
         raise TypeError(f'Layout for {type(box).__name__} not handled yet')
     return (
@@ -768,7 +782,7 @@ def split_inline_box(context, box, position_x, max_x, bottom_space, skip_stack,
 
             if new_position_x > max_x and not trailing_whitespace:
                 previous_resume_at = _break_waiting_children(
-                    context, box, max_x, bottom_space, initial_skip_stack,
+                    context, containing_block, max_x, bottom_space, initial_skip_stack,
                     absolute_boxes, fixed_boxes, line_placeholders,
                     waiting_floats, line_children, children, waiting_children)
                 if previous_resume_at:
@@ -844,7 +858,7 @@ def split_inline_box(context, box, position_x, max_x, bottom_space, skip_stack,
                 skip_stack=None)
 
     if resume_at is not None:
-        index = tuple(resume_at)[0]
+        index = next(iter(resume_at))
         if index < float_resume_index:
             resume_at = {float_resume_index: None}
 
@@ -871,28 +885,16 @@ def split_text_box(context, box, available_width, skip, is_line_start=True):
     """
     assert isinstance(box, boxes.TextBox)
     font_size = box.style['font_size']
-    text = box.text[skip:]
+    text = box.text.encode()[skip:]
     if font_size == 0 or not text:
         return None, None, False
     layout, length, resume_index, width, height, baseline = split_first_line(
-        text, box.style, context, available_width, box.justification_spacing,
-        is_line_start=is_line_start)
+        text.decode(), box.style, context, available_width,
+        box.justification_spacing, is_line_start=is_line_start)
     assert resume_index != 0
 
-    # Convert ``length`` and ``resume_at`` from UTF-8 indexes in text
-    # to Unicode indexes.
-    # No need to encode what’s after resume_at (if set) or length (if
-    # resume_at is not set). One code point is one or more byte, so
-    # UTF-8 indexes are always bigger or equal to Unicode indexes.
-    new_text = layout.text
-    encoded = text.encode()
-    if resume_index is not None:
-        between = encoded[length:resume_index].decode()
-        resume_index = len(encoded[:resume_index].decode())
-    length = len(encoded[:length].decode())
-
     if length > 0:
-        box = box.copy_with_text(new_text)
+        box = box.copy_with_text(layout.text)
         box.width = width
         box.pango_layout = layout
         # "The height of the content area should be based on the font,
@@ -917,6 +919,7 @@ def split_text_box(context, box, available_width, skip, is_line_start=True):
     if resume_index is None:
         preserved_line_break = False
     else:
+        between = text[length:resume_index].decode()
         preserved_line_break = (
             (length != resume_index) and between.strip(' '))
         if preserved_line_break:
@@ -1046,7 +1049,9 @@ def inline_box_verticality(box, top_bottom_subtrees, baseline_y):
 
         # the child’s `top` is `child.baseline` above (lower y) its baseline.
         top = child_baseline_y - child.baseline
-        if isinstance(child, (boxes.InlineBlockBox, boxes.InlineFlexBox)):
+        box_types = (
+            boxes.InlineBlockBox, boxes.InlineFlexBox, boxes.InlineGridBox)
+        if isinstance(child, box_types):
             # This also includes table wrappers for inline tables.
             child.translate(dy=top - child.position_y)
         else:
