@@ -81,6 +81,7 @@ DEF_TYPES = frozenset((
     'mask',
     'path',
     'pattern',
+    'symbol',
 ))
 
 
@@ -130,69 +131,68 @@ class Node:
         """Whether node is visible."""
         return self.display and self.get('visibility') != 'hidden'
 
+    def cascade(self, child):
+        """Apply CSS cascade and other related operations to given child."""
+        wrapper = child._wrapper
+
+        # Cascade
+        for key, value in self.attrib.items():
+            if key not in NOT_INHERITED_ATTRIBUTES:
+                if key not in child.attrib:
+                    child.attrib[key] = value
+
+        # Apply style attribute
+        if style_attr := child.get('style'):
+            normal_attr, important_attr = parse_declarations(style_attr)
+        else:
+            normal_attr, important_attr = [], []
+        normal_matcher, important_matcher = self._style
+        normal = [rule[-1] for rule in normal_matcher.match(wrapper)]
+        important = [rule[-1] for rule in important_matcher.match(wrapper)]
+        for declarations_list in (normal, [normal_attr], important, [important_attr]):
+            for declarations in declarations_list:
+                for name, value in declarations:
+                    child.attrib[name] = value.strip()
+
+        # Replace 'currentColor' value
+        for key in COLOR_ATTRIBUTES:
+            if child.get(key) == 'currentColor':
+                child.attrib[key] = child.get('color', 'black')
+
+        # Handle 'inherit' values
+        for key, value in child.attrib.copy().items():
+            if value == 'inherit':
+                value = self.get(key)
+                if value is None:
+                    del child.attrib[key]
+                else:
+                    child.attrib[key] = value
+
+        # Fix text in text tags
+        if child.tag in ('text', 'textPath', 'a'):
+            children, _ = child.text_children(
+                wrapper, trailing_space=True, text_root=True)
+            child._wrapper.etree_children = [
+                child._etree_node for child in children]
+
+
     def __iter__(self):
         """Yield node children, handling cascade."""
         for wrapper in self._wrapper:
             child = Node(wrapper, self._style)
-
-            # Cascade
-            for key, value in self.attrib.items():
-                if key not in NOT_INHERITED_ATTRIBUTES:
-                    if key not in child.attrib:
-                        child.attrib[key] = value
-
-            # Apply style attribute
-            style_attr = child.get('style')
-            if style_attr:
-                normal_attr, important_attr = parse_declarations(style_attr)
-            else:
-                normal_attr, important_attr = [], []
-            normal_matcher, important_matcher = self._style
-            normal = [rule[-1] for rule in normal_matcher.match(wrapper)]
-            important = [rule[-1] for rule in important_matcher.match(wrapper)]
-            declarations_lists = (
-                normal, [normal_attr], important, [important_attr])
-            for declarations_list in declarations_lists:
-                for declarations in declarations_list:
-                    for name, value in declarations:
-                        child.attrib[name] = value.strip()
-
-            # Replace 'currentColor' value
-            for key in COLOR_ATTRIBUTES:
-                if child.get(key) == 'currentColor':
-                    child.attrib[key] = child.get('color', 'black')
-
-            # Handle 'inherit' values
-            for key, value in child.attrib.copy().items():
-                if value == 'inherit':
-                    value = self.get(key)
-                    if value is None:
-                        del child.attrib[key]
-                    else:
-                        child.attrib[key] = value
-
-            # Fix text in text tags
-            if child.tag in ('text', 'textPath', 'a'):
-                children, _ = child.text_children(
-                    wrapper, trailing_space=True, text_root=True)
-                child._wrapper.etree_children = [
-                    child._etree_node for child in children]
-
+            self.cascade(child)
             yield child
 
     def get_viewbox(self):
         """Get node viewBox as a tuple of floats."""
         viewbox = self.get('viewBox')
         if viewbox:
-            return tuple(
-                float(number) for number in normalize(viewbox).split())
+            return tuple(float(number) for number in normalize(viewbox).split())
 
     def get_href(self, base_url):
         """Get the href attribute, with or without a namespace."""
         for attr_name in ('{http://www.w3.org/1999/xlink}href', 'href'):
-            url = get_url_attribute(
-                self, attr_name, base_url, allow_relative=True)
-            if url:
+            if url := get_url_attribute(self, attr_name, base_url, allow_relative=True):
                 return url
 
     def del_href(self):
@@ -307,6 +307,19 @@ class Node:
         self.__class__ = type(
             'Node', (Node,), {'__iter__': lambda _: iterator})
 
+    def set_svg_size(self, svg, concrete_width, concrete_height):
+        """"Set SVG concrete and inner widths and heights from svg node."""
+        svg.concrete_width = concrete_width
+        svg.concrete_height = concrete_height
+        svg.normalized_diagonal = hypot(concrete_width, concrete_height) / sqrt(2)
+
+        if viewbox := self.get_viewbox():
+            svg.inner_width, svg.inner_height = viewbox[2], viewbox[3]
+        else:
+            svg.inner_width, svg.inner_height = svg.concrete_width, svg.concrete_height
+        svg.inner_diagonal = hypot(svg.inner_width, svg.inner_height) / sqrt(2)
+
+
 
 class SVG:
     """An SVG document."""
@@ -316,12 +329,6 @@ class SVG:
         style = parse_stylesheets(wrapper, url)
         self.tree = Node(wrapper, style)
         self.url = url
-
-        # Replace 'currentColor' value
-        for key in COLOR_ATTRIBUTES:
-            if self.tree.get(key) == 'currentColor':
-                self.tree.attrib[key] = self.tree.get('color', 'black')
-
         self.filters = {}
         self.gradients = {}
         self.images = {}
@@ -329,6 +336,7 @@ class SVG:
         self.masks = {}
         self.patterns = {}
         self.paths = {}
+        self.symbols = {}
 
         self.use_cache = {}
 
@@ -336,6 +344,7 @@ class SVG:
         self.cursor_d_position = [0, 0]
         self.text_path_width = 0
 
+        self.tree.cascade(self.tree)
         self.parse_defs(self.tree)
         self.inherit_defs()
 
@@ -374,19 +383,7 @@ class SVG:
         """Draw image on a stream."""
         self.stream = stream
 
-        self.concrete_width = concrete_width
-        self.concrete_height = concrete_height
-        self.normalized_diagonal = (
-            hypot(concrete_width, concrete_height) / sqrt(2))
-
-        viewbox = self.get_viewbox()
-        if viewbox:
-            self.inner_width, self.inner_height = viewbox[2], viewbox[3]
-        else:
-            self.inner_width = self.concrete_width
-            self.inner_height = self.concrete_height
-        self.inner_diagonal = (
-            hypot(self.inner_width, self.inner_height) / sqrt(2))
+        self.tree.set_svg_size(self, concrete_width, concrete_height)
 
         self.base_url = base_url
         self.url_fetcher = url_fetcher
@@ -447,21 +444,21 @@ class SVG:
             if new_ctm.determinant:
                 self.stream.transform(*(old_ctm @ new_ctm.invert).values)
 
-        # Handle text anchor
-        if node.tag == 'text':
-            text_anchor = node.get('text-anchor')
-            children = tuple(node)
-            if children and not node.text:
-                text_anchor = children[0].get('text-anchor')
-        if node.tag == 'text' and text_anchor in ('middle', 'end'):
-            group = self.stream.add_group(0, 0, 0, 0)  # BBox set after drawing
-            original_streams.append(self.stream)
-            self.stream = group
-
-        # Set text bounding box
+        # Handle text anchor and set text bounding box
+        text_anchor_shift = False
         if node.display and TAGS.get(node.tag) == text:
+            if (text_anchor := node.get('text-anchor')) in ('middle', 'end'):
+                text_anchor_shift = True
+                group = self.stream.add_group(0, 0, 0, 0)  # BBox set after drawing
+                original_streams.append(self.stream)
+                self.stream = group
             node.text_bounding_box = EMPTY_BOUNDING_BOX
 
+        # Save concrete size of root svg tag
+        if node.tag == 'svg':
+            concrete_width = self.concrete_width
+            concrete_height = self.concrete_height
+
         # Draw node
         if node.visible and node.tag in TAGS:
             with suppress(PointError):
@@ -470,7 +467,12 @@ class SVG:
         # Draw node children
         if node.display and node.tag not in DEF_TYPES:
             for child in node:
+                if text_anchor_shift:
+                    new_stream = self.stream
+                    self.stream = original_streams[-1]
                 self.draw_node(child, font_size, fill_stroke)
+                if text_anchor_shift:
+                    self.stream = new_stream
                 visible_text_child = (
                     TAGS.get(node.tag) == text and
                     TAGS.get(child.tag) == text and
@@ -484,8 +486,12 @@ class SVG:
                     node.text_bounding_box = extend_bounding_box(
                         node.text_bounding_box, ((x1, y1), (x2, y2)))
 
+        # Restore concrete and inner size of root svg tag
+        if node.tag == 'svg':
+            self.tree.set_svg_size(svg, concrete_width, concrete_height)
+
         # Handle text anchor
-        if node.tag == 'text' and text_anchor in ('middle', 'end'):
+        if text_anchor_shift:
             group_id = self.stream.id
             self.stream = original_streams.pop()
             self.stream.push_state()
@@ -560,13 +566,10 @@ class SVG:
                 position = 'end'
 
             # Draw marker
-            marker = markers[position]
-            if not marker:
+            if not (marker_node := self.markers.get(markers[position])):
                 position = 'mid' if angles else 'start'
                 continue
 
-            marker_node = self.markers.get(marker)
-
             # Calculate position, scale and clipping
             translate_x, translate_y = self.point(
                 marker_node.get('refX'), marker_node.get('refY'),
@@ -670,9 +673,9 @@ class SVG:
         fill_drawn = draw_gradient_or_pattern(
             self, node, fill_source, font_size, fill_opacity, stroke=False)
         if fill_color and not fill_drawn:
-            red, green, blue, alpha = color(fill_color)
-            self.stream.set_color_rgb(red, green, blue)
-            self.stream.set_alpha(alpha * fill_opacity)
+            stream_color = color(fill_color)
+            stream_color.alpha *= fill_opacity
+            self.stream.set_color(stream_color)
         fill = fill_color or fill_drawn
 
         # Get stroke data
@@ -681,9 +684,9 @@ class SVG:
         stroke_drawn = draw_gradient_or_pattern(
             self, node, stroke_source, font_size, stroke_opacity, stroke=True)
         if stroke_color and not stroke_drawn:
-            red, green, blue, alpha = color(stroke_color)
-            self.stream.set_color_rgb(red, green, blue, stroke=True)
-            self.stream.set_alpha(alpha * stroke_opacity, stroke=True)
+            stream_color = color(stroke_color)
+            stream_color.alpha *= stroke_opacity
+            self.stream.set_color(stream_color, stroke=True)
         stroke = stroke_color or stroke_drawn
         stroke_width = self.length(node.get('stroke-width', '1px'), font_size)
         if stroke_width:
