@@ -4,19 +4,12 @@ from itertools import cycle
 from math import ceil, hypot
 
 from ..matrix import Matrix
-from .bounding_box import is_valid_bounding_box
-from .utils import color, parse_url, size, transform
+from .bounding_box import bounding_box, is_valid_bounding_box
+from .utils import alpha_value, color, parse_url, size, transform
 
 
-def use(svg, node, font_size):
-    """Draw use tags."""
-    from . import NOT_INHERITED_ATTRIBUTES, SVG
-
-    x, y = svg.point(node.get('x'), node.get('y'), font_size)
-
-    for attribute in ('x', 'y', 'viewBox', 'mask'):
-        if attribute in node.attrib:
-            del node.attrib[attribute]
+def get_use_tree(svg, node, font_size):
+    from . import SVG
 
     parsed_url = parse_url(node.get_href(svg.url))
     svg_url = parse_url(svg.url)
@@ -46,24 +39,37 @@ def use(svg, node, font_size):
             use_svg.get_intrinsic_size(font_size)
             tree = use_svg.tree
 
+    return tree
+
+
+def use(svg, node, font_size):
+    """Draw use tags."""
+    x, y = svg.point(node.get('x'), node.get('y'), font_size)
+
+    for attribute in ('x', 'y', 'viewBox', 'mask'):
+        if attribute in node.attrib:
+            del node.attrib[attribute]
+
+    if (tree := get_use_tree(svg, node, font_size)) is None:
+        return
+
     if tree.tag in ('svg', 'symbol'):
         # Explicitely specified
         # https://www.w3.org/TR/SVG11/struct.html#UseElement
-        tree._etree_node.tag = 'svg'
         if 'width' in node.attrib and 'height' in node.attrib:
             tree.attrib['width'] = node.attrib['width']
             tree.attrib['height'] = node.attrib['height']
+        else:
+            tree._etree_node.tag = 'g'
+            box = bounding_box(svg, tree, font_size, stroke=True)
+            if is_valid_bounding_box(box):
+                tree.attrib['width'] = box[0] + box[2]
+                tree.attrib['height'] = box[1] + box[3]
+        tree._etree_node.tag = 'svg'
 
-    # Cascade
-    for key, value in node.attrib.items():
-        if key not in NOT_INHERITED_ATTRIBUTES:
-            if key not in tree.attrib:
-                tree.attrib[key] = value
-
-    svg.stream.push_state()
+    node.cascade(tree)
+    node.override_iter(iter((tree,)))
     svg.stream.transform(e=x, f=y)
-    svg.draw_node(tree, font_size)
-    svg.stream.pop_state()
 
 
 def draw_gradient_or_pattern(svg, node, name, font_size, opacity, stroke):
@@ -85,20 +91,15 @@ def draw_gradient(svg, node, gradient, font_size, opacity, stroke):
         positions.append(max(
             positions[-1] if positions else 0,
             size(child.get('offset'), font_size, 1)))
-        stop_opacity = float(child.get('stop-opacity', 1)) * opacity
+        stop_opacity = alpha_value(child.get('stop-opacity', 1)) * opacity
         stop_color = color(child.get('stop-color', 'black'))
-        if stop_opacity < 1:
-            stop_color = tuple(
-                stop_color[:3] + (stop_color[3] * stop_opacity,))
+        stop_color.alpha *= stop_opacity
         colors.append(stop_color)
 
     if not colors:
         return False
     elif len(colors) == 1:
-        red, green, blue, alpha = colors[0]
-        svg.stream.set_color_rgb(red, green, blue)
-        if alpha != 1:
-            svg.stream.set_alpha(alpha, stroke=stroke)
+        svg.stream.set_color(colors[0])
         return True
 
     bounding_box = svg.calculate_bounding_box(node, font_size, stroke)
@@ -190,6 +191,14 @@ def draw_gradient(svg, node, gradient, font_size, opacity, stroke):
         bx2, by2 = transform_matrix.invert.transform_point(width, height)
         width, height = bx2 - bx1, by2 - by1
 
+        # Ensure that width and height are positive to please some PDF readers
+        if bx1 > bx2:
+            width = -width
+            bx1, bx2 = bx2, bx1
+        if by1 > by2:
+            height = -height
+            by1, by2 = by2, by1
+
     pattern = svg.stream.add_pattern(
         bx1, by1, width, height, width, height, matrix @ svg.stream.ctm)
     group = pattern.add_group(bx1, by1, width, height)
@@ -221,10 +230,10 @@ def draw_gradient(svg, node, gradient, font_size, opacity, stroke):
             shading_type, 'Gray', domain, coords, extend, function)
         alpha_stream.stream = [f'/{alpha_shading.id} sh']
 
-    group.shading(shading.id)
+    group.paint_shading(shading.id)
     pattern.set_alpha(1)
     pattern.draw_x_object(group.id)
-    svg.stream.color_space('Pattern', stroke=stroke)
+    svg.stream.set_color_space('Pattern', stroke=stroke)
     svg.stream.set_color_special(pattern.id, stroke=stroke)
     return True
 
@@ -251,17 +260,15 @@ def spread_linear_gradient(spread, positions, colors, x1, y1, x2, y2,
 
         # Create cycles used to add colors
         if spread == 'repeat':
-            next_steps = cycle([0] + position_steps)
+            next_steps = cycle((0, *position_steps))
             next_colors = cycle(colors)
-            previous_steps = cycle([0] + position_steps[::-1])
+            previous_steps = cycle((0, *position_steps[::-1]))
             previous_colors = cycle(colors[::-1])
         else:
             assert spread == 'reflect'
-            next_steps = cycle(
-                [0] + position_steps[::-1] + [0] + position_steps)
+            next_steps = cycle((0, *position_steps[::-1], 0, *position_steps))
             next_colors = cycle(colors[::-1] + colors)
-            previous_steps = cycle(
-                [0] + position_steps + [0] + position_steps[::-1])
+            previous_steps = cycle((0, *position_steps, 0, *position_steps[::-1]))
             previous_colors = cycle(colors + colors[::-1])
 
         # Normalize bounding box
@@ -404,8 +411,7 @@ def spread_radial_gradient(spread, positions, colors, fx, fy, fr, cx, cy, r,
                 new_positions = [
                     position - 1 - full_repeat for position
                     in original_positions[-(i - 1):]]
-                positions = (
-                    [ratio - 1 - full_repeat] + new_positions + positions)
+                positions = [ratio - 1 - full_repeat, *new_positions, *positions]
                 break
 
     coords = (fx, fy, fr, cx, cy, r)
@@ -424,18 +430,10 @@ def draw_pattern(svg, node, pattern, font_size, opacity, stroke):
     x, y = bounding_box[0], bounding_box[1]
     matrix = Matrix(e=x, f=y)
     if pattern.get('patternUnits') == 'userSpaceOnUse':
-        width, height = svg.inner_width, svg.inner_height
-    else:
-        width, height = bounding_box[2], bounding_box[3]
-
-    if pattern.get('patternUnits') == 'userSpaceOnUse':
-        x = size(pattern.get('x'), font_size, 1)
-        y = size(pattern.get('y'), font_size, 1)
         pattern_width = size(pattern.get('width', 0), font_size, 1)
         pattern_height = size(pattern.get('height', 0), font_size, 1)
     else:
-        x = size(pattern.get('x'), font_size, 1) * width
-        y = size(pattern.get('y'), font_size, 1) * height
+        width, height = bounding_box[2], bounding_box[3]
         pattern_width = (
             size(pattern.attrib.pop('width', '1'), font_size, 1) * width)
         pattern_height = (
@@ -466,7 +464,7 @@ def draw_pattern(svg, node, pattern, font_size, opacity, stroke):
         group, pattern_width, pattern_height, svg.base_url,
         svg.url_fetcher, svg.context)
     stream_pattern.draw_x_object(group.id)
-    svg.stream.color_space('Pattern', stroke=stroke)
+    svg.stream.set_color_space('Pattern', stroke=stroke)
     svg.stream.set_color_special(stream_pattern.id, stroke=stroke)
     return True
 
