@@ -5,7 +5,7 @@ from contextlib import suppress
 from math import cos, radians, sin, tan
 from urllib.parse import urlparse
 
-from tinycss2.color3 import parse_color
+from tinycss2.color4 import parse_color
 
 from ..matrix import Matrix
 
@@ -57,6 +57,17 @@ def size(string, font_size=None, percentage_reference=None):
     return 0
 
 
+def alpha_value(value):
+    """Return opacity between 0 and 1 from str, number or percentage."""
+    ratio = 1
+    if isinstance(value, str):
+        value = value.strip()
+        if value.endswith('%'):
+            ratio = 100
+            value = value[:-1].strip()
+    return min(1, max(0, float(value) / ratio))
+
+
 def point(svg, string, font_size):
     """Pop first two size values from a string."""
     match = re.match('(.*?) (.*?)(?: |$)', string)
@@ -135,7 +146,7 @@ def parse_url(url):
 
 def color(string):
     """Safely parse a color string and return a RGBA tuple."""
-    return parse_color(string or '') or (0, 0, 0, 1)
+    return parse_color(string or '') or parse_color('black')
 
 
 def transform(transform_string, font_size, normalized_diagonal):
