@@ -1,6 +1,7 @@
 """Imports of dynamic libraries used for text layout."""
 
 import os
+import sys
 from contextlib import suppress
 
 import cffi
@@ -12,20 +13,69 @@ ffi.cdef('''
     typedef ... hb_font_t;
     typedef ... hb_face_t;
     typedef ... hb_blob_t;
+    typedef int hb_bool_t;
+    typedef uint32_t hb_tag_t;
     typedef uint32_t hb_codepoint_t;
-    hb_face_t * hb_font_get_face (hb_font_t *font);
+    hb_tag_t hb_tag_from_string (const char *str, int len);
+    void hb_tag_to_string (hb_tag_t tag, char *buf);
+    void hb_face_destroy (hb_face_t *face);
     hb_blob_t * hb_face_reference_blob (hb_face_t *face);
     unsigned int hb_face_get_index (const hb_face_t *face);
     unsigned int hb_face_get_upem (const hb_face_t *face);
+    hb_blob_t * hb_face_reference_table (const hb_face_t *face, hb_tag_t tag);
     const char * hb_blob_get_data (hb_blob_t *blob, unsigned int *length);
+    unsigned int hb_blob_get_length (hb_blob_t *blob);
     bool hb_ot_color_has_png (hb_face_t *face);
-    hb_blob_t * hb_ot_color_glyph_reference_png (
-        hb_font_t *font, hb_codepoint_t glyph);
+    hb_blob_t * hb_ot_color_glyph_reference_png (hb_font_t *font, hb_codepoint_t glyph);
     bool hb_ot_color_has_svg (hb_face_t *face);
-    hb_blob_t * hb_ot_color_glyph_reference_svg (
-        hb_face_t *face, hb_codepoint_t glyph);
+    hb_blob_t * hb_ot_color_glyph_reference_svg (hb_face_t *face, hb_codepoint_t glyph);
     void hb_blob_destroy (hb_blob_t *blob);
+    unsigned int hb_face_get_table_tags (
+        const hb_face_t *face, unsigned int start_offset, unsigned int *table_count,
+        hb_tag_t *table_tags);
+    hb_bool_t hb_version_atleast (
+        unsigned int major, unsigned int minor, unsigned int micro);
 
+    // HarfBuzz Subset
+
+    typedef ... hb_subset_input_t;
+    typedef ... hb_set_t;
+
+    typedef enum {
+        HB_SUBSET_FLAGS_DEFAULT = 0x00000000u,
+        HB_SUBSET_FLAGS_NO_HINTING = 0x00000001u,
+        HB_SUBSET_FLAGS_RETAIN_GIDS = 0x00000002u,
+        HB_SUBSET_FLAGS_DESUBROUTINIZE = 0x00000004u,
+        HB_SUBSET_FLAGS_NAME_LEGACY = 0x00000008u,
+        HB_SUBSET_FLAGS_SET_OVERLAPS_FLAG = 0x00000010u,
+        HB_SUBSET_FLAGS_PASSTHROUGH_UNRECOGNIZED = 0x00000020u,
+        HB_SUBSET_FLAGS_NOTDEF_OUTLINE = 0x00000040u,
+        HB_SUBSET_FLAGS_GLYPH_NAMES = 0x00000080u,
+        HB_SUBSET_FLAGS_NO_PRUNE_UNICODE_RANGES = 0x00000100u,
+        HB_SUBSET_FLAGS_NO_LAYOUT_CLOSURE = 0x00000200u,
+    } hb_subset_flags_t;
+
+    typedef enum {
+        HB_SUBSET_SETS_GLYPH_INDEX = 0,
+        HB_SUBSET_SETS_UNICODE,
+        HB_SUBSET_SETS_NO_SUBSET_TABLE_TAG,
+        HB_SUBSET_SETS_DROP_TABLE_TAG,
+        HB_SUBSET_SETS_NAME_ID,
+        HB_SUBSET_SETS_NAME_LANG_ID,
+        HB_SUBSET_SETS_LAYOUT_FEATURE_TAG,
+        HB_SUBSET_SETS_LAYOUT_SCRIPT_TAG,
+    } hb_subset_sets_t;
+
+    hb_subset_input_t * hb_subset_input_create_or_fail (void);
+    void hb_subset_input_destroy (hb_subset_input_t *input);
+    hb_set_t * hb_subset_input_glyph_set (hb_subset_input_t *input);
+    void hb_set_add_sorted_array (
+        hb_set_t *set, const hb_codepoint_t *sorted_codepoints,
+        unsigned int num_codepoints);
+    hb_face_t * hb_subset_or_fail (hb_face_t *source, const hb_subset_input_t *input);
+    void hb_subset_input_set_flags (hb_subset_input_t *input, unsigned  value);
+    hb_set_t * hb_subset_input_set (
+        hb_subset_input_t *input, hb_subset_sets_t set_type);
 
     // Pango
 
@@ -95,6 +145,16 @@ ffi.cdef('''
         PANGO_WRAP_WORD_CHAR
     } PangoWrapMode;
 
+    typedef enum {
+        PANGO_VARIANT_NORMAL,
+        PANGO_VARIANT_SMALL_CAPS,
+        PANGO_VARIANT_ALL_SMALL_CAPS,
+        PANGO_VARIANT_PETITE_CAPS,
+        PANGO_VARIANT_ALL_PETITE_CAPS,
+        PANGO_VARIANT_UNICASE,
+        PANGO_VARIANT_TITLE_CAPS,
+    } PangoVariant;
+
     typedef enum {
         PANGO_TAB_LEFT
     } PangoTabAlign;
@@ -209,29 +269,27 @@ ffi.cdef('''
 
     PangoLayout * pango_layout_new (PangoContext *context);
     void pango_layout_set_width (PangoLayout *layout, int width);
-    PangoAttrList * pango_layout_get_attributes(PangoLayout *layout);
-    void pango_layout_set_attributes (
-        PangoLayout *layout, PangoAttrList *attrs);
-    void pango_layout_set_text (
-        PangoLayout *layout, const char *text, int length);
-    void pango_layout_set_tabs (
-        PangoLayout *layout, PangoTabArray *tabs);
+    PangoAttrList * pango_layout_get_attributes (PangoLayout *layout);
+    void pango_layout_set_attributes (PangoLayout *layout, PangoAttrList *attrs);
+    void pango_layout_set_text (PangoLayout *layout, const char *text, int length);
+    void pango_layout_set_tabs (PangoLayout *layout, PangoTabArray *tabs);
     void pango_layout_set_font_description (
         PangoLayout *layout, const PangoFontDescription *desc);
-    void pango_layout_set_wrap (
-        PangoLayout *layout, PangoWrapMode wrap);
-    void pango_layout_set_single_paragraph_mode (
-        PangoLayout *layout, gboolean setting);
+    void pango_layout_set_wrap (PangoLayout *layout, PangoWrapMode wrap);
+    void pango_layout_set_single_paragraph_mode (PangoLayout *layout, gboolean setting);
+    void pango_layout_set_ellipsize (PangoLayout *layout, PangoEllipsizeMode ellipsize);
     int pango_layout_get_baseline (PangoLayout *layout);
-    PangoLayoutLine * pango_layout_get_line_readonly (
-        PangoLayout *layout, int line);
+    void pango_layout_line_get_extents (
+        PangoLayoutLine *line, PangoRectangle *ink_rect, PangoRectangle *logical_rect);
+    PangoLayoutLine * pango_layout_get_line_readonly (PangoLayout *layout, int line);
+    const PangoLogAttr* pango_layout_get_log_attrs_readonly (
+        PangoLayout* layout, gint* n_attrs);
 
     hb_font_t * pango_font_get_hb_font (PangoFont *font);
 
     PangoFontDescription * pango_font_description_new (void);
     void pango_font_description_free (PangoFontDescription *desc);
-    PangoFontDescription * pango_font_description_copy (
-        const PangoFontDescription *desc);
+    PangoFontMap* pango_font_get_font_map (PangoFont* font);
 
     void pango_font_description_set_family (
         PangoFontDescription *desc, const char *family);
@@ -245,33 +303,27 @@ ffi.cdef('''
         PangoFontDescription *desc, double size);
     void pango_font_description_set_variations (
         PangoFontDescription* desc, const char* variations);
+    void pango_font_description_set_variant (
+        PangoFontDescription* desc, PangoVariant variant);
 
-    PangoStyle pango_font_description_get_style (
-        const PangoFontDescription *desc);
+    PangoStyle pango_font_description_get_style (const PangoFontDescription *desc);
     const char* pango_font_description_get_variations (
         const PangoFontDescription* desc);
-    PangoWeight pango_font_description_get_weight (
-        const PangoFontDescription* desc);
+    PangoWeight pango_font_description_get_weight (const PangoFontDescription* desc);
     int pango_font_description_get_size (PangoFontDescription *desc);
 
     void pango_font_description_unset_fields (
         PangoFontDescription* desc, PangoFontMask to_unset);
 
-    int pango_glyph_string_get_width (PangoGlyphString *glyphs);
-    char * pango_font_description_to_string (
-        const PangoFontDescription *desc);
+    char * pango_font_description_to_string (const PangoFontDescription *desc);
 
     PangoFontDescription * pango_font_describe (PangoFont *font);
-    const char * pango_font_description_get_family (
-        const PangoFontDescription *desc);
+    const char * pango_font_description_get_family (const PangoFontDescription *desc);
     guint pango_font_description_hash (const PangoFontDescription *desc);
 
-    PangoContext * pango_context_new ();
     PangoContext * pango_font_map_create_context (PangoFontMap *fontmap);
-
     PangoFont* pango_font_map_load_font (
-        PangoFontMap* fontmap, PangoContext* context,
-        const PangoFontDescription* desc);
+        PangoFontMap* fontmap, PangoContext* context, const PangoFontDescription* desc);
 
     PangoFontMetrics * pango_context_get_metrics (
         PangoContext *context, const PangoFontDescription *desc,
@@ -281,14 +333,10 @@ ffi.cdef('''
     void pango_font_metrics_unref (PangoFontMetrics *metrics);
     int pango_font_metrics_get_ascent (PangoFontMetrics *metrics);
     int pango_font_metrics_get_descent (PangoFontMetrics *metrics);
-    int pango_font_metrics_get_underline_thickness (
-        PangoFontMetrics *metrics);
-    int pango_font_metrics_get_underline_position (
-        PangoFontMetrics *metrics);
-    int pango_font_metrics_get_strikethrough_thickness (
-        PangoFontMetrics *metrics);
-    int pango_font_metrics_get_strikethrough_position (
-        PangoFontMetrics *metrics);
+    int pango_font_metrics_get_underline_thickness (PangoFontMetrics *metrics);
+    int pango_font_metrics_get_underline_position (PangoFontMetrics *metrics);
+    int pango_font_metrics_get_strikethrough_thickness (PangoFontMetrics *metrics);
+    int pango_font_metrics_get_strikethrough_position (PangoFontMetrics *metrics);
     void pango_font_get_glyph_extents (
         PangoFont *font, PangoGlyph glyph, PangoRectangle *ink_rect,
         PangoRectangle *logical_rect);
@@ -298,14 +346,11 @@ ffi.cdef('''
 
     PangoAttrList * pango_attr_list_new (void);
     void pango_attr_list_unref (PangoAttrList *list);
-    void pango_attr_list_insert (
-        PangoAttrList *list, PangoAttribute *attr);
-    void pango_attr_list_change (
-        PangoAttrList *list, PangoAttribute *attr);
+    void pango_attr_list_insert (PangoAttrList *list, PangoAttribute *attr);
+    void pango_attr_list_change (PangoAttrList *list, PangoAttribute *attr);
     PangoAttribute * pango_attr_font_features_new (const gchar *features);
     PangoAttribute * pango_attr_letter_spacing_new (int letter_spacing);
     PangoAttribute * pango_attr_insert_hyphens_new (gboolean insert_hyphens);
-    void pango_attribute_destroy (PangoAttribute *attr);
 
     PangoTabArray * pango_tab_array_new_with_positions (
         gint size, gboolean positions_in_pixels, PangoTabAlign first_alignment,
@@ -314,19 +359,7 @@ ffi.cdef('''
 
     PangoLanguage * pango_language_from_string (const char *language);
     PangoLanguage * pango_language_get_default (void);
-    void pango_context_set_language (
-        PangoContext *context, PangoLanguage *language);
-    void pango_context_set_font_map (
-        PangoContext *context, PangoFontMap *font_map);
-
-    void pango_layout_line_get_extents (
-        PangoLayoutLine *line,
-        PangoRectangle *ink_rect, PangoRectangle *logical_rect);
-
-    PangoContext * pango_layout_get_context (PangoLayout *layout);
-    void pango_layout_set_ellipsize (
-        PangoLayout *layout,
-        PangoEllipsizeMode ellipsize);
+    void pango_context_set_language (PangoContext *context, PangoLanguage *language);
 
     void pango_get_log_attrs (
         const char *text, int length, int level, PangoLanguage *language,
@@ -363,47 +396,48 @@ ffi.cdef('''
 
     FcConfig * FcInitLoadConfigAndFonts (void);
     void FcConfigDestroy (FcConfig *config);
-    FcBool FcConfigAppFontAddFile (
-        FcConfig *config, const FcChar8 *file);
-    FcConfig * FcConfigGetCurrent (void);
-    FcBool FcConfigSetCurrent (FcConfig *config);
-    FcBool FcConfigParseAndLoad (
-        FcConfig *config, const FcChar8 *file, FcBool complain);
+    FcBool FcConfigAppFontAddFile (FcConfig *config, const FcChar8 *file);
+    FcBool FcConfigParseAndLoadFromMemory (
+        FcConfig *config, const FcChar8 *buffer, FcBool complain);
 
-    FcFontSet * FcConfigGetFonts(FcConfig *config, FcSetName set);
-    FcStrList * FcConfigGetConfigFiles(FcConfig *config);
-    FcChar8 * FcStrListNext(FcStrList *list);
+    FcFontSet * FcConfigGetFonts (FcConfig *config, FcSetName set);
+    FcStrList * FcConfigGetConfigFiles (FcConfig *config);
+    FcChar8 * FcStrListNext (FcStrList *list);
 
     void FcDefaultSubstitute (FcPattern *pattern);
-    FcBool FcConfigSubstitute (
-        FcConfig *config, FcPattern *p, FcMatchKind kind);
+    FcBool FcConfigSubstitute (FcConfig *config, FcPattern *p, FcMatchKind kind);
 
     FcPattern * FcPatternCreate (void);
     FcPattern * FcPatternDestroy (FcPattern *p);
-    FcBool FcPatternAddString (
-        FcPattern *p, const char *object, const FcChar8 *s);
-    FcResult FcPatternGetString (
-        FcPattern *p, const char *object, int n, FcChar8 **s);
-    FcPattern * FcFontMatch (
-        FcConfig *config, FcPattern *p, FcResult *result);
+    FcBool FcPatternAddString (FcPattern *p, const char *object, const FcChar8 *s);
+    FcResult FcPatternGetString (FcPattern *p, const char *object, int n, FcChar8 **s);
+    FcPattern * FcFontMatch (FcConfig *config, FcPattern *p, FcResult *result);
 
 
     // PangoFT2
 
+    typedef ... PangoFcFont;
     typedef ... PangoFcFontMap;
 
     PangoFontMap * pango_ft2_font_map_new (void);
-    void pango_fc_font_map_set_config (
-        PangoFcFontMap *fcfontmap, FcConfig *fcconfig);
+    void pango_fc_font_map_set_config (PangoFcFontMap *fcfontmap, FcConfig *fcconfig);
     void pango_fc_font_map_config_changed (PangoFcFontMap *fcfontmap);
+    hb_face_t* pango_fc_font_map_get_hb_face (
+         PangoFcFontMap* fcfontmap, PangoFcFont* fcfont);
 ''')
 
 
-def _dlopen(ffi, *names):
+def _dlopen(ffi, *names, allow_fail=False):
     """Try various names for the same library, for different platforms."""
+    if os.name == 'nt':
+        flags = 0x00001000  # LOAD_LIBRARY_SEARCH_DEFAULT_DIRS
+    else:
+        flags = ffi.RTLD_NOW  # default
     for name in names:
         with suppress(OSError):
-            return ffi.dlopen(name)
+            return ffi.dlopen(name, flags)
+    if allow_fail:
+        return
     # Re-raise the exception.
     print(
         '\n-----\n\n'
@@ -414,38 +448,43 @@ def _dlopen(ffi, *names):
         'https://doc.courtbouillon.org/weasyprint/stable/'
         'first_steps.html#troubleshooting',
         '\n\n-----\n')  # pragma: no cover
-    return ffi.dlopen(names[0])  # pragma: no cover
+    return ffi.dlopen(names[0], flags)  # pragma: no cover
 
 
-if hasattr(os, 'add_dll_directory'):  # pragma: no cover
+if hasattr(os, 'add_dll_directory') and not hasattr(sys, 'frozen'):  # pragma: no cover
     dll_directories = os.getenv(
         'WEASYPRINT_DLL_DIRECTORIES',
+        'C:\\msys64\\mingw64\\bin;'
         'C:\\Program Files\\GTK3-Runtime Win64\\bin').split(';')
     for dll_directory in dll_directories:
         with suppress((OSError, FileNotFoundError)):
             os.add_dll_directory(dll_directory)
 
 gobject = _dlopen(
-    ffi, 'gobject-2.0-0', 'gobject-2.0', 'libgobject-2.0-0',
-    'libgobject-2.0.so.0', 'libgobject-2.0.dylib',  'libgobject-2.0-0.dll')
+    ffi, 'libgobject-2.0-0', 'gobject-2.0-0', 'gobject-2.0',
+    'libgobject-2.0.so.0', 'libgobject-2.0.dylib', 'libgobject-2.0-0.dll')
 pango = _dlopen(
-    ffi, 'pango-1.0-0', 'pango-1.0', 'libpango-1.0-0', 'libpango-1.0.so.0',
+    ffi, 'libpango-1.0-0', 'pango-1.0-0', 'pango-1.0', 'libpango-1.0.so.0',
     'libpango-1.0.dylib', 'libpango-1.0-0.dll')
 harfbuzz = _dlopen(
-    ffi, 'harfbuzz', 'harfbuzz-0.0', 'libharfbuzz-0',
-    'libharfbuzz.so.0', 'libharfbuzz.so.0', 'libharfbuzz.0.dylib',
-    'libharfbuzz-0.dll')
+    ffi, 'libharfbuzz-0', 'harfbuzz', 'harfbuzz-0.0',
+    'libharfbuzz.so.0', 'libharfbuzz.0.dylib', 'libharfbuzz-0.dll')
+harfbuzz_subset = _dlopen(
+    ffi, 'libharfbuzz-subset-0', 'harfbuzz-subset', 'harfbuzz-subset-0.0',
+    'libharfbuzz-subset.so.0', 'libharfbuzz-subset.0.dylib', 'libharfbuzz-subset-0.dll',
+    allow_fail=True)
 fontconfig = _dlopen(
-    ffi, 'fontconfig-1', 'fontconfig', 'libfontconfig', 'libfontconfig.so.1',
-    'libfontconfig.1.dylib', 'libfontconfig-1.dll')
+    ffi, 'libfontconfig-1', 'fontconfig-1', 'fontconfig',
+    'libfontconfig.so.1', 'libfontconfig.1.dylib', 'libfontconfig-1.dll')
 pangoft2 = _dlopen(
-    ffi, 'pangoft2-1.0-0', 'pangoft2-1.0', 'libpangoft2-1.0-0',
+    ffi, 'libpangoft2-1.0-0', 'pangoft2-1.0-0', 'pangoft2-1.0',
     'libpangoft2-1.0.so.0', 'libpangoft2-1.0.dylib', 'libpangoft2-1.0-0.dll')
 
 gobject.g_type_init()
 
-units_to_double = pango.pango_units_to_double
-units_from_double = pango.pango_units_from_double
+# Call once to avoid int overflows.
+TO_UNITS = pango.pango_units_from_double(1)
+FROM_UNITS = pango.pango_units_to_double(1)
 
 
 def unicode_to_char_p(string):
