@@ -6,9 +6,7 @@ from math import inf
 import pyphen
 
 from .constants import LST_TO_ISO, PANGO_WRAP_MODE
-from .ffi import (
-    ffi, gobject, pango, pangoft2, unicode_to_char_p, units_from_double,
-    units_to_double)
+from .ffi import FROM_UNITS, TO_UNITS, ffi, gobject, pango, pangoft2, unicode_to_char_p
 from .fonts import font_features, get_font_description
 
 
@@ -20,8 +18,8 @@ def line_size(line, style):
     """
     logical_extents = ffi.new('PangoRectangle *')
     pango.pango_layout_line_get_extents(line, ffi.NULL, logical_extents)
-    width = units_to_double(logical_extents.width)
-    height = units_to_double(logical_extents.height)
+    width = logical_extents.width * FROM_UNITS
+    height = logical_extents.height * FROM_UNITS
     ffi.release(logical_extents)
     if style['letter_spacing'] != 'normal':
         width += style['letter_spacing']
@@ -51,7 +49,7 @@ def first_line_metrics(first_line, text, layout, resume_at, space_collapse,
         length = first_line.length if first_line is not None else 0
 
     width, height = line_size(first_line, style)
-    baseline = units_to_double(pango.pango_layout_get_baseline(layout.layout))
+    baseline = pango.pango_layout_get_baseline(layout.layout) * FROM_UNITS
     layout.deactivate()
     return layout, length, resume_at, width, height, baseline
 
@@ -106,15 +104,15 @@ class Layout:
                 pango.pango_context_get_metrics(
                     pango_context, font_description, self.language),
                 pango.pango_font_metrics_unref)
-            self.ascent = units_to_double(
+            self.ascent = FROM_UNITS * (
                 pango.pango_font_metrics_get_ascent(metrics))
-            self.underline_position = units_to_double(
+            self.underline_position = FROM_UNITS * (
                 pango.pango_font_metrics_get_underline_position(metrics))
-            self.strikethrough_position = units_to_double(
+            self.strikethrough_position = FROM_UNITS * (
                 pango.pango_font_metrics_get_strikethrough_position(metrics))
-            self.underline_thickness = units_to_double(
+            self.underline_thickness = FROM_UNITS * (
                 pango.pango_font_metrics_get_underline_thickness(metrics))
-            self.strikethrough_thickness = units_to_double(
+            self.strikethrough_thickness = FROM_UNITS * (
                 pango.pango_font_metrics_get_strikethrough_thickness(metrics))
         else:
             self.ascent = None
@@ -178,15 +176,23 @@ class Layout:
                 pango.pango_attr_list_change(attr_list, attr)
 
             if letter_spacing:
-                letter_spacing = units_from_double(letter_spacing)
+                letter_spacing = int(letter_spacing * TO_UNITS)
                 add_attr(0, len(bytestring), letter_spacing)
 
             if word_spacing:
-                space_spacing = (
-                    units_from_double(word_spacing) + letter_spacing)
+                if bytestring == b' ':
+                    # We need more than one space to set word spacing
+                    self.text = ' \u200b'  # Space + zero-width space
+                    text, bytestring = unicode_to_char_p(self.text)
+                    pango.pango_layout_set_text(self.layout, text, -1)
+
+                space_spacing = int(word_spacing * TO_UNITS + letter_spacing)
                 position = bytestring.find(b' ')
+                # Pango gives only half of word-spacing on boundaries
+                boundary_positions = (0, len(bytestring) - 1)
                 while position != -1:
-                    add_attr(position, position + 1, space_spacing)
+                    factor = 1 + (position in boundary_positions)
+                    add_attr(position, position + 1, factor * space_spacing)
                     position = bytestring.find(b' ', position + 1)
 
             if word_breaking:
@@ -226,23 +232,14 @@ class Layout:
 
 
 def create_layout(text, style, context, max_width, justification_spacing):
-    """Return an opaque Pango layout with default Pango line-breaks.
-
-    :param text: Unicode
-    :param style: a style dict of computed values
-    :param max_width:
-        The maximum available width in the same unit as ``style['font_size']``,
-        or ``None`` for unlimited width.
-
-    """
+    """Return an opaque Pango layout with default Pango line-breaks."""
     layout = Layout(context, style, justification_spacing, max_width)
 
     # Make sure that max_width * Pango.SCALE == max_width * 1024 fits in a
     # signed integer. Treat bigger values same as None: unconstrained width.
     text_wrap = style['white_space'] in ('normal', 'pre-wrap', 'pre-line')
     if max_width is not None and text_wrap and max_width < 2 ** 21:
-        pango.pango_layout_set_width(
-            layout.layout, units_from_double(max(0, max_width)))
+        pango.pango_layout_set_width(layout.layout, int(max(0, max_width) * TO_UNITS))
 
     layout.set_text(text)
     return layout
@@ -273,81 +270,101 @@ def split_first_line(text, style, context, max_width, justification_spacing,
     if not text_wrap:
         max_width = None
 
-    # Step #1: Get a draft layout with the first line
+    # Step #1: Get a draft layout with the first line.
+    ratio = 4  # number that almost always respects char_height / char_width > ratio
+    short_text = text
     if max_width is not None and max_width != inf and style['font_size']:
-        short_text = text
-        if max_width == 0:
-            # Trying to find minimum size, let's naively split on spaces and
-            # keep one word + one letter
+        # Try to use a small amount of text to avoid the whole layout. We need
+        # at least one line, and one possible line break point on the second line.
+        if style['font_size'] * ratio > max_width:
+            # Trying to find minimum or very small size, let's naively split on
+            # spaces and keep one word + one letter.
             space_index = text.find(' ')
             if space_index != -1:
                 short_text = text[:space_index+2]  # index + space + one letter
         else:
-            short_text = text[:int(max_width / style['font_size'] * 2.5)]
-        # Try to use a small amount of text instead of the whole text
+            # Use the magic ration and hope that we’ll get the right amount of text.
+            short_text = text[:int(max_width / style['font_size'] * ratio)]
         layout = create_layout(
             short_text, style, context, max_width, justification_spacing)
         first_line, resume_index = layout.get_first_line()
         if resume_index is None and short_text != text:
-            # The small amount of text fits in one line, give up and use
-            # the whole text
+            # The small amount of text fits in one line, give up and use the
+            # whole text.
+            short_text = text
             layout.set_text(text)
             first_line, resume_index = layout.get_first_line()
+        else:
+            # If the second line of the short text can break, we have the next
+            # line break point required for step #3 in it, drop the end of the text.
+            first_line_text = short_text.encode()[:resume_index].decode()
+            if first_line_text != short_text:
+                start, end = len(first_line_text) + 1, len(short_text)
+                text_end_log_attrs = pango.pango_layout_get_log_attrs_readonly(
+                    layout.layout, ffi.NULL)[start:end]
+                if get_next_break_point(text_end_log_attrs) is not None:
+                    text = short_text
     else:
         layout = create_layout(
             text, style, context, original_max_width, justification_spacing)
         first_line, resume_index = layout.get_first_line()
 
-    # Step #2: Don't split lines when it's not needed
+    # Step #2: Don't split lines when it's not needed.
     if max_width is None:
-        # The first line can take all the place needed
+        # The first line can take all the place needed.
         return first_line_metrics(
             first_line, text, layout, resume_index, space_collapse, style)
     first_line_width, _ = line_size(first_line, style)
     if resume_index is None and first_line_width <= max_width:
-        # The first line fits in the available width
+        # The first line fits in the available width.
         return first_line_metrics(
             first_line, text, layout, resume_index, space_collapse, style)
 
     # Step #3: Try to put the first word of the second line on the first line
     # https://mail.gnome.org/archives/gtk-i18n-list/2013-September/msg00006
     # is a good thread related to this problem.
-    first_line_text = text.encode()[:resume_index].decode()
-    first_line_fits = (
-        first_line_width <= max_width or
-        ' ' in first_line_text.strip() or
-        can_break_text(first_line_text.strip(), style['lang']))
-    if first_line_fits:
-        # The first line fits but may have been cut too early by Pango
-        second_line_text = text.encode()[resume_index:].decode()
+    if first_line_width <= max_width:
+        # The first line fits but may have been cut too early by Pango.
+        encoded_text = text.encode()
+        first_line_text = encoded_text[:resume_index].decode()
+        second_line_text = encoded_text[resume_index:].decode()
     else:
         # The line can't be split earlier, try to hyphenate the first word.
         first_line_text = ''
         second_line_text = text
-
-    next_word = second_line_text.split(' ', 1)[0]
+    if first_line_text == short_text:
+        # There’s no second line, don’t try to find a next word.
+        break_point = None
+    else:
+        # Find then second line’s first break point.
+        log_attrs = pango.pango_layout_get_log_attrs_readonly(layout.layout, ffi.NULL)
+        start, end = len(first_line_text) + 1, len(short_text)
+        second_line_log_attrs = log_attrs[start:end]
+        break_point = get_next_break_point(second_line_log_attrs)
+        if break_point is not None:
+            break_point -= len(first_line_text) + 1
+    next_word = second_line_text[:break_point].rstrip(' ')
     if next_word:
-        if space_collapse:
-            # next_word might fit without a space afterwards
-            # only try when space collapsing is allowed
+        if space_collapse and second_line_text[break_point or -1] == ' ':
+            # Next word might fit without a space afterwards only try when
+            # space collapsing is allowed.
             new_first_line_text = first_line_text + next_word
             layout.set_text(new_first_line_text)
             first_line, resume_index = layout.get_first_line()
             if resume_index is None:
                 if first_line_text:
-                    # The next word fits in the first line, keep the layout
+                    # The next word fits in the first line, keep the layout.
                     resume_index = len(new_first_line_text.encode()) + 1
                     return first_line_metrics(
-                        first_line, text, layout, resume_index, space_collapse,
-                        style)
+                        first_line, text, layout, resume_index, space_collapse, style)
                 else:
-                    # Second line is None
+                    # Second line is None.
                     resume_index = first_line.length + 1
                     if resume_index >= len(text.encode()):
                         resume_index = None
     elif first_line_text:
         # We found something on the first line but we did not find a word on
-        # the next line, no need to hyphenate, we can keep the current layout
+        # the next line, no need to hyphenate, we can keep the current layout.
         return first_line_metrics(
             first_line, text, layout, resume_index, space_collapse, style)
 
@@ -360,7 +377,7 @@ def split_first_line(text, style, context, max_width, justification_spacing,
 
     auto_hyphenation = manual_hyphenation = False
     if hyphens != 'none':
-        manual_hyphenation = soft_hyphen in first_line_text + next_word
+        manual_hyphenation = soft_hyphen in first_line_text + second_line_text
     if hyphens == 'auto' and lang:
         next_word_boundaries = get_next_word_boundaries(second_line_text, lang)
         if next_word_boundaries:
@@ -390,19 +407,11 @@ def split_first_line(text, style, context, max_width, justification_spacing,
         # hyphen and add the missing hyphen
         if first_line_text.endswith(soft_hyphen):
             # The first line has been split on a soft hyphen
-            if ' ' in first_line_text:
-                first_line_text, next_word = first_line_text.rsplit(' ', 1)
-                next_word = f' {next_word}'
-                layout.set_text(first_line_text)
-                first_line, _ = layout.get_first_line()
-                resume_index = len((f'{first_line_text} ').encode())
-            else:
-                first_line_text, next_word = '', first_line_text
+            first_line_text, second_line_text = '', first_line_text
         soft_hyphen_indexes = [
-            match.start() for match in re.finditer(soft_hyphen, next_word)]
+            match.start() for match in re.finditer(soft_hyphen, second_line_text)]
         soft_hyphen_indexes.reverse()
-        dictionary_iterations = [
-            next_word[:i + 1] for i in soft_hyphen_indexes]
+        dictionary_iterations = [second_line_text[:i+1] for i in soft_hyphen_indexes]
         start_word = 0
     elif auto_hyphenation:
         dictionary_key = (lang, left, right, total)
@@ -429,19 +438,12 @@ def split_first_line(text, style, context, max_width, justification_spacing,
             new_first_line, index = new_layout.get_first_line()
             new_first_line_width, _ = line_size(new_first_line, style)
             new_space = max_width - new_first_line_width
-            hyphenated = (
-                index is None and (
-                    new_space >= 0 or
-                    first_word_part == dictionary_iterations[-1]))
+            hyphenated = index is None and (
+                new_space >= 0 or first_word_part == dictionary_iterations[-1])
             if hyphenated:
                 layout = new_layout
                 first_line = new_first_line
                 resume_index = len(new_first_line_text.encode())
-                if text[len(new_first_line_text)] == soft_hyphen:
-                    # Recreate the layout with no max_width to be sure that
-                    # we don't break before the soft hyphen
-                    pango.pango_layout_set_width(layout.layout, -1)
-                    resume_index += len(soft_hyphen.encode())
                 break
 
         if not hyphenated and not first_line_text:
@@ -486,10 +488,8 @@ def split_first_line(text, style, context, max_width, justification_spacing,
         # instance) from keeping their shape when wrapped on the next line with
         # pango layout. Maybe insert Unicode shaping characters in text?
         layout.set_text(text)
-        pango.pango_layout_set_width(
-            layout.layout, units_from_double(max_width))
-        pango.pango_layout_set_wrap(
-            layout.layout, PANGO_WRAP_MODE['WRAP_CHAR'])
+        pango.pango_layout_set_width(layout.layout, int(max_width * TO_UNITS))
+        pango.pango_layout_set_wrap(layout.layout, PANGO_WRAP_MODE['WRAP_CHAR'])
         first_line, index = layout.get_first_line()
         resume_index = index or first_line.length
         if resume_index >= len(text.encode()):
@@ -516,21 +516,27 @@ def get_log_attrs(text, lang):
     log_attrs = ffi.new('PangoLogAttr[]', length)
     pango.pango_get_log_attrs(
         text_p, len(bytestring), -1, language, log_attrs, length)
-    return bytestring, log_attrs
+    return log_attrs
+
+
+def get_next_break_point(log_attrs):
+    for i, attr in enumerate(log_attrs):
+        if attr.is_line_break:
+            return i
 
 
 def can_break_text(text, lang):
     if not text or len(text) < 2:
         return None
-    bytestring, log_attrs = get_log_attrs(text, lang)
+    log_attrs = get_log_attrs(text, lang)
     length = len(text) + 1
-    return any(attr.is_line_break for attr in log_attrs[1:length - 1])
+    return get_next_break_point(log_attrs[1:length-1]) is not None
 
 
 def get_next_word_boundaries(text, lang):
     if not text or len(text) < 2:
         return None
-    bytestring, log_attrs = get_log_attrs(text, lang)
+    log_attrs = get_log_attrs(text, lang)
     for i, attr in enumerate(log_attrs):
         if attr.is_word_end:
             word_end = i
@@ -545,7 +551,7 @@ def get_next_word_boundaries(text, lang):
 def get_last_word_end(text, lang):
     if not text or len(text) < 2:
         return None
-    bytestring, log_attrs = get_log_attrs(text, lang)
+    log_attrs = get_log_attrs(text, lang)
     for i, attr in enumerate(list(log_attrs)[::-1]):
         if i and attr.is_word_end:
             return len(text) - i
